/*======================================================================

    PCMCIA Card Manager

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <syslog.h>
#include <getopt.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/wait.h>

#include "cs_types.h"
#include "cs.h"
#include "cistpl.h"
#include "ds.h"

#include "cardmgr.h"

static const char *version =
    "cardmgr.c v$Revision: 1.11 $: $Date: 1994/06/22 23:19:28 $ (David Hinds)";

/*====================================================================*/

typedef struct socket_info_t {
    int fd;
    int state;
    card_info_t *card;
    bind_info_t bind[MAX_FUNCTIONS];
} socket_info_t;

#define SOCKET_PRESENT	0x01
#define SOCKET_READY	0x02
#define SOCKET_BOUND	0x04

/* Linked list of resource adjustments */
struct adjust_list_t *root_adjust = NULL;

/* Linked list of device definitions */
struct device_info_t *root_device = NULL;

/* Linked list of card definitions */
struct card_info_t *root_card = NULL;

static struct socket_info_t socket[MAX_SOCKS];

/* Default path for config file */
static char *config = "/etc/pcmcia/config";
/* Default path for finding modules */
static char *modpath = "/usr/src/linux/modules";

/*====================================================================*/

card_info_t *lookup_card(int ns)
{
    socket_info_t *s = &socket[ns];
    card_info_t *card;
    ds_ioctl_arg_t arg;
    cistpl_vers_1_t *vers;
    int i, ret, match;

    /* Try to read VERS_1 tuple */
    vers = NULL;
    arg.tuple.TupleData = arg.tuple_parse.data;
    arg.tuple.TupleDataMax = sizeof(arg.tuple_parse.data);
    arg.tuple.Attributes = 0;
    arg.tuple.DesiredTuple = CISTPL_VERS_1;
    do {
	ret = ioctl(s->fd, DS_GET_FIRST_TUPLE, &arg);
	if (ret != 0)
	    break;
	arg.tuple.TupleOffset = 0;
	ret = ioctl(s->fd, DS_GET_TUPLE_DATA, &arg);
	if (ret != 0) {
	    syslog(LOG_ERR, "error reading CIS data on socket %d", ns);
	    break;
	}
	ret = ioctl(s->fd, DS_PARSE_TUPLE, &arg);
	if (ret != 0) {
	    syslog(LOG_ERR, "error parsing CIS on socket %d", ns);
	    break;
	}
	vers = &arg.tuple_parse.parse.version_1;
    } while (0);

    match = 0;
    for (card = root_card; card; card = card->next) {
	switch (card->ident_type) {
	    
	case VERS_1_IDENT:
	    if ((vers == NULL) || (card->vers.ns != vers->ns))
		break;
	    for (i = 0; i < vers->ns; i++) {
		if (strcmp(card->vers.pi[i], "*") == 0)
		    continue;
		if (strcmp(card->vers.pi[i], vers->pi[i]) != 0)
		    break;
	    }
	    if (i < vers->ns)
		break;
	    match = 1;
	    break;
	    
	case TUPLE_IDENT:
	    arg.tuple.DesiredTuple = card->tuple.code;
	    ret = ioctl(s->fd, DS_GET_FIRST_TUPLE, &arg);
	    if (ret != 0) break;
	    arg.tuple.TupleOffset = card->tuple.ofs;
	    ret = ioctl(s->fd, DS_GET_TUPLE_DATA, &arg);
	    if (ret != 0) break;
	    if (strncmp(arg.tuple_parse.data, card->tuple.info,
			strlen(card->tuple.info)) != 0)
		break;
	    match = 1;
	    break;
	    
	}
	if (match) break;
    }
    if (match)
	syslog(LOG_INFO, "socket %d: %s", ns, card->name);
    else {
	syslog(LOG_INFO, "unsupported card in socket %d", ns);
	if (vers && (vers->ns > 0)) {
	    char v[128] = "";
	    for (i = 0; i < vers->ns; i++)
		sprintf(v+strlen(v), "%s\"%s\"",
			(i>0) ? ", " : "", vers->pi[i]);
	    syslog(LOG_INFO, "version info: %s", v);
	}
    }
    return card;
} /* lookup_card */

/*====================================================================*/

static int execute(char *msg, char *cmd)
{
    int ret;
    syslog(LOG_INFO, "executing: '%s'", cmd);
    ret = system(cmd);
    if (WIFEXITED(ret)) {
	if (WEXITSTATUS(ret))
	    syslog(LOG_ERR, "%s exited with status %d",
		   msg, WEXITSTATUS(ret));
	return WEXITSTATUS(ret);
    }
    else {
	syslog(LOG_ERR, "%s exited on signal %d",
	       msg, WTERMSIG(ret));
    }
    return -1;
} /* execute */

/*====================================================================*/

static int execute_on_dev(char *msg, char *fmt, char *dev)
{
    int ret, ofs;
    char *cmd, *s;
	
    s = strstr(fmt, "%d%");
    if (s == NULL)
	return execute(msg, fmt);
    cmd = malloc(strlen(fmt)+strlen(dev)-2);
    ofs = s-fmt;
    strncpy(cmd, fmt, ofs);
    strcpy(cmd+ofs, dev);
    strcat(cmd, s+3);
    
    syslog(LOG_INFO, "executing: '%s'", cmd);
    ret = system(cmd);
    if (WIFEXITED(ret)) {
	if (WEXITSTATUS(ret))
	    syslog(LOG_ERR, "%s exited with status %d",
		   msg, WEXITSTATUS(ret));
	return WEXITSTATUS(ret);
    }
    else {
	syslog(LOG_ERR, "%s exited on signal %d",
	       msg, WTERMSIG(ret));
    }
    return -1;
} /* execute_on_dev */

/*====================================================================*/

static void do_reset(int i)
{
    socket_info_t *s = &socket[i];
    int ret;
    
    if (s->state & (SOCKET_PRESENT|SOCKET_READY)) {
	syslog(LOG_ERR, "bogus insertion event on socket %d", i);
	return;
    }
    s->state |= SOCKET_PRESENT;
    ret = ioctl(s->fd, DS_RESET_CARD, NULL);
    if (ret != 0) {
	syslog(LOG_ERR, "reset socket %d failed: %m", i);
	return;
    }
}

/*====================================================================*/

static void do_insert(int sn)
{
    socket_info_t *s = &socket[sn];
    card_info_t *card;
    device_info_t **dev;
    bind_info_t *bind;
    int i, j, ret;
    char cmd[128];

    if (s->state & SOCKET_READY) {
	syslog(LOG_ERR, "bogus reset event on socket %d", sn);
	return;
    }
    s->state |= SOCKET_READY;
    
    s->card = card = lookup_card(sn);
    if (card == NULL)
	return;

    /* Initialize socket bindings */
    dev = card->device; bind = s->bind;
    for (i = 0; i < card->functions; i++) {
	strcpy((char *)bind[i].dev_info, (char *)dev[i]->dev_info);
	bind[i].instance = NULL;
	bind[i].name[0] = '\0';
    }
    
    /* Install kernel modules */
    for (i = 0; i < card->functions; i++) {
	ret = 0;
	for (j = 0; j < dev[i]->modules; j++) {
	    sprintf(cmd, "insmod %s/%s.o", modpath, dev[i]->module[j]);
	    ret = execute("insmod", cmd);
	    if ((ret != 0) && (ret != 2)) break;
	}
	if ((ret != 0) && (ret != 2))
	    return;
    }
    
    /* Bind drivers by their dev_info identifiers */
    for (i = 0; i < card->functions; i++) {
	ret = ioctl(s->fd, DS_BIND_REQUEST, &bind[i]);
	if (ret != 0) {
	    syslog(LOG_ERR, "bind '%s' to socket %d failed: %m",
		   (char *)bind[i].dev_info, sn);
	    if (errno != EBUSY) return;
	}
	for (j = 0; j < 10; j++) {
	    ret = ioctl(s->fd, DS_GET_DEVICE_INFO, &bind[i]);
	    if ((ret == 0) || (errno != EAGAIN))
		break;
	    usleep(100000);
	}
	if (ret != 0) {
	    syslog(LOG_ERR, "get dev info on socket %d failed: %m",
		   sn);
	    return;
	}
	if (strlen(bind[i].name) == 0) {
	    syslog(LOG_ERR, "no dev info for socket %d", sn);
	    return;
	}
    }
    
    /* Run "start" commands */
    for (i = 0; i < card->functions; i++) {
	if (dev[i]->start_cmd)
	    execute_on_dev("start cmd", dev[i]->start_cmd, bind[i].name);
    }
} /* do_insert */

/*====================================================================*/

void do_remove(int sn)
{
    socket_info_t *s = &socket[sn];
    card_info_t *card;
    device_info_t **dev;
    bind_info_t *bind;
    int i, j, ret;
    char cmd[128];
    
    if (!((s->state & SOCKET_PRESENT) &&
	  (s->state & SOCKET_READY))) {
	syslog(LOG_INFO, "do_remove() on uninitialized socket");
	return;
    }
    s->state &= ~(SOCKET_PRESENT | SOCKET_READY);
    
    card = s->card;
    if (card == NULL)
	return;
    
    /* Run "stop" commands */
    dev = card->device; bind = s->bind;
    for (i = 0; i < card->functions; i++) {
	if (dev[i]->stop_cmd && (bind[i].name[0] != '\0'))
	    execute_on_dev("stop cmd", dev[i]->stop_cmd, bind[i].name);
    }

    /* unbind driver instances */
    for (i = 0; i < card->functions; i++) {
	if (bind[i].instance != NULL) {
	    ret = ioctl(s->fd, DS_UNBIND_REQUEST, &s->bind[i]);
	    if (ret != 0)
		syslog(LOG_ERR, "unbind '%s' from socket %d failed: %m",
		       (char *)bind[i].dev_info, sn);
	}
    }

    /* remove kernel modules in inverse order */
    for (i = 0; i < card->functions; i++) {
	for (j = dev[i]->modules-1; j >= 0; j--) {
	    sprintf(cmd, "rmmod %s", dev[i]->module[j]);
	    execute("rmmod", cmd);
	}
    }
}

/*====================================================================*/

void adjust_resources(int fd)
{
    adjust_list_t *al;
    int ret;
    char tmp[64];
    
    for (al = root_adjust; al; al = al->next) {
	ret = ioctl(fd, DS_ADJUST_RESOURCE_INFO, &al->adj);
	if (ret != 0) {
	    switch (al->adj.Resource) {
	    case RES_MEMORY_RANGE:
		sprintf(tmp, "memory 0x%p-0x%p",
			al->adj.resource.memory.Base,
			al->adj.resource.memory.Base +
			al->adj.resource.memory.Size - 1);
		break;
	    case RES_IO_RANGE:
		sprintf(tmp, "IO ports %#x-%#x",
			al->adj.resource.io.BasePort,
			al->adj.resource.io.BasePort +
			al->adj.resource.io.NumPorts - 1);
		break;
	    case RES_IRQ:
		sprintf(tmp, "irq %d", al->adj.resource.irq.IRQ);
		break;
	    }
	    syslog(LOG_ERR, "could not adjust resource: %s: %m", tmp);
	}
    }
} /* adjust_resources */
    
/*====================================================================*/

void done(void)
{
    syslog(LOG_INFO, "exiting");
}

void main(int argc, char *argv[])
{
    int optch, errflg, sockets, fd;
    int i, ret, event;
    extern char *optarg;
    fd_set fds;
    char fn[20];
    FILE *f;

    errflg = 0;
    while ((optch = getopt(argc, argv, "c:m:")) != -1) {
	switch (optch) {
	case 'c':
	    config = strdup(optarg); break;
	case 'm':
	    modpath = strdup(optarg); break;
	default:
	    errflg = 1; break;
	}
    }
    if (errflg || (optind < argc)) {
	fprintf(stderr, "usage: %s [-c config]\n", argv[0]);
	exit(EXIT_FAILURE);
    }

#ifdef DEBUG
    openlog("cardmgr", LOG_PID|LOG_PERROR, LOG_DAEMON);
#else
    openlog("cardmgr", LOG_PID, LOG_DAEMON);
#endif

    syslog(LOG_INFO, "starting");
    syslog(LOG_INFO, "%s", version);
    atexit(&done);

#ifndef DEBUG
    if ((ret = fork()) > 0) exit(0);
    if (ret == -1)
	syslog(LOG_ERR, "forking: %m");
    if (setsid() < 0)
	syslog(LOG_ERR, "detaching from tty: %m");
#endif
    
    f = fopen(config, "r");
    if (f == NULL) {
	syslog(LOG_ERR, "could not open %s: %m", config);
	exit(EXIT_FAILURE);
    }
    parse_configfile(f);
    fclose(f);

    for (i = 0; i < MAX_SOCKS; i++) {
	sprintf(fn, "/dev/pc%d", i);
	fd = open(fn, O_RDWR);
	if (fd < 0) break;
	socket[i].fd = fd;
	socket[i].state = 0;
    }
    sockets = i;
    if (sockets == 0) {
	syslog(LOG_ERR, "no sockets found!");
	exit(EXIT_FAILURE);
    }
    else
	syslog(LOG_INFO, "watching %d sockets", sockets);

    adjust_resources(socket[0].fd);
    
    for (;;) {
	FD_ZERO(&fds);
	for (i = 0; i < sockets; i++)
	    FD_SET(socket[i].fd, &fds);
	
	while ((ret = select(MAX_SOCKS+4, &fds, NULL, NULL, NULL)) < 0) {
	    syslog(LOG_ERR, "select(): %m");
	    exit(EXIT_FAILURE);
	}

	for (i = 0; i < sockets; i++) {
	    if (!FD_ISSET(socket[i].fd, &fds))
		continue;
	    ret = read(socket[i].fd, &event, 4);
	    if ((ret == -1) && (errno != EAGAIN))
		syslog(LOG_ERR, "read(): %m\n");
	    if (ret == 4) {
		if (event & DS_EVENT_CARD_REMOVAL)
		    do_remove(i);
		if (event & DS_EVENT_CARD_INSERTION)
		    do_reset(i);
		if (event & DS_EVENT_RESET_COMPLETE)
		    do_insert(i);
	    }
	}
    } /* repeat */
}
