/*======================================================================

    PCMCIA Card Information Structure parser

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/major.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/sched.h>

#include "cs_types.h"
#include "cs.h"
#include "cistpl.h"

/*======================================================================

    Parsing routines for individual tuples
    
======================================================================*/

#define MIN(a, b) (((a) < (b)) ? (a) : (b))

static int parse_vers_1(tuple_t *tuple, cistpl_vers_1_t *vers_1)
{
    int i, len, sl;
    char *p;
    
    len = MIN(tuple->TupleDataLen, tuple->TupleDataMax);
    if (len < 2) return CS_BAD_TUPLE;
    vers_1->major = tuple->TupleData[0];
    vers_1->minor = tuple->TupleData[1];
    len -= 2;
    p = (char *)tuple->TupleData + 2;
    for (i = 0; i < CISTPL_VERS_1_MAX_PROD_STRINGS; i++) {
	if (len < 2) break;
	/* !!! this might not be safe !!! */
	sl = strlen(p) + 1;
	if ((sl > CISTPL_VERS_1_STRLEN) || (sl > len))
	    return CS_BAD_TUPLE;
	strcpy(vers_1->pi[i], p);
	len -= sl; p += sl;
    }
    vers_1->ns = i;
    return CS_SUCCESS;
} /* parse_vers_1 */

/*====================================================================*/

#define CISTPL_CONFIG_ADDR_MASK 0x03
#define CISTPL_CONFIG_RMASK_MASK 0x3c

static int parse_config(tuple_t *tuple, cistpl_config_t *config)
{
    int rasz, rmsz, i;
    u_char *p;

    rasz = tuple->TupleData[0] & CISTPL_CONFIG_ADDR_MASK;
    rmsz = (tuple->TupleData[0] & CISTPL_CONFIG_RMASK_MASK) >> 2;
    if (tuple->TupleDataLen < rasz+rmsz+4)
	return CS_BAD_TUPLE;
    config->last_idx = tuple->TupleData[1];
    p = tuple->TupleData + 2;
    config->base = 0;
    for (i = 0; i <= rasz; i++)
	config->base += p[i] << (8*i);
    p += rasz+1;
    for (i = 0; i < 4; i++)
	config->rmask[i] = 0;
    for (i = 0; i < rmsz; i++)
	config->rmask[i>>2] += p[i] << (8*(i%4));
    return CS_SUCCESS;
} /* parse_config */

/*====================================================================*/

int parse_tuple(client_handle_t handle, tuple_t *tuple,
		cisparse_t *parse)
{
    int ret = CS_SUCCESS;
    if (tuple->TupleDataLen > tuple->TupleDataMax)
	return CS_BAD_TUPLE;
    switch (tuple->TupleCode) {
    case CISTPL_VERS_1:
	ret = parse_vers_1(tuple, &parse->version_1);
	break;
    case CISTPL_CONFIG:
	ret = parse_config(tuple, &parse->config);
	break;
    default:
	ret = CS_UNSUPPORTED_FUNCTION;
	break;
    }
    return ret;
} /* parse_tuple */
