/*======================================================================

    PCMCIA Card Services driver

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/major.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <asm/system.h>

#ifdef MODULE
#include <linux/module.h>
#include "../../tools/version.h"
#endif

#include "cs_types.h"
/* Throw away prototype for CardServices() */
#define CardServices XCardServices
#include "cs.h"
#undef CardServices
#include "cistpl.h"
#include "cisreg.h"

#include <linux/pcmcia.h>

#define CHECK_HANDLE(h) \
    (((h) == NULL) || ((h)->client_magic != CLIENT_MAGIC))

#define CHECK_SOCKET(s) \
    (((s) >= sockets) || (socket_table[s].ctrl == NULL))
#define SOCKET(h) (&socket_table[(h)->Socket])

#define EVENT(h, e, p) \
    ((h)->event_handler((e), (p), &(h)->event_callback_args));

#ifdef PCMCIA_DEBUG
int pc_debug = PCMCIA_DEBUG;
#endif

static const char *version =
    "cs.c v$Revision: 1.27 $: $Date: 1994/06/24 16:23:10 $ (David Hinds)\n";

/*====================================================================*/

typedef struct irq_info_t {
    u_long Attributes;
    int time_share, dyn_share;
} irq_info_t;

/* Possible irq uses */
#define CSC_IRQ	1
#define IO_IRQ	2

typedef struct socket_info_t {
    pcmcia_ctrl_t *ctrl;
    u_long state;
    client_handle_t clients;
    client_handle_t reset_handle;
    struct timer_list setup, shutdown, unreset;
    int unreset_timeout;
    struct pcmcia_mem_map attr_mem;
    resource_entry_t attr_mem_alloc;
    u_long Attributes;
    long Vcc, Vpp1, Vpp2;
    long IntType;
    caddr_t ConfigBase;
    u_char Status, Pin, Copy, Option;
    long Present;
    long FirstDevType, FuncCode;
    u_long SysInitMask;
    long ManufCode, ManufInfo;
    long CardValues;
    irq_req_t irq;
    io_req_t io;
} socket_info_t;

/* Flags in socket state */
#define SOCKET_CONFIG_LOCKED	0x01
#define SOCKET_IRQ_REQ		0x02
#define SOCKET_IO_REQ		0x04
#define SOCKET_PRESENT		0x08
#define SOCKET_SETUP_PENDING	0x10
#define SOCKET_SHUTDOWN_PENDING	0x20
#define SOCKET_RESET_PENDING	0x40

/*====================================================================*/

#define MAX_SOCK 8

/* Table of sockets */
static socket_t sockets = 0;
static socket_info_t socket_table[MAX_SOCK];

/* Memory resource database */
static resource_entry_t mem_db = { 0, 0, &mem_db };

/* IO port resource database */
static resource_entry_t io_db = { 0, 0, &io_db };

/* Table of IRQ assignments */
static irq_info_t irq_table[16] = { { 0, 0, 0 }, /* etc */ };

/*======================================================================

    Very basic initialization stuff -- if we're linked into the base
    kernel, this will be invoked from init.c; otherwise, it gets
    called at module initialization time.
    
======================================================================*/

ulong cs_init(u_long kmem_start, u_long kmem_end)
{
#ifdef CONFIG_I82365
    i365_init();
#endif
#ifdef CONFIG_TCIC
    tcic_init();
#endif
    printk("PCMCIA Card Services level 2.10\n");
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk(version);
#endif
    return kmem_start;
} /* cs_init */

/*======================================================================

    Setting a card's COR seems to need a short delay.  We have to
    busy loop because it might happen in an interrupt routine.  So
    here's a generic busy loop.
    
======================================================================*/

static void busy_loop(u_long len)
{
    u_long timeout = jiffies + len;
    u_long flags;
    save_flags(flags);
    sti();
    while (timeout >= jiffies)
	;
    restore_flags(flags);
} /* busy_loop */
  
/*======================================================================

    These manage the internal databases of available resources
    
======================================================================*/

static int add_interval(resource_entry_t *map, u_long base, u_long num)
{
    resource_entry_t *p, *q;

    for (p = map; ; p = p->next) {
	if ((p != map) && (p->base+p->num-1 >= base))
	    return CS_IN_USE;
	if ((p->next == map) || (p->next->base > base+num-1))
	    break;
    }
    q = kmalloc(sizeof(resource_entry_t), GFP_KERNEL);
    q->base = base; q->num = num;
    q->next = p->next; p->next = q;
    return CS_SUCCESS;
} /* add_interval */

/*====================================================================*/

static int sub_interval(resource_entry_t *map, u_long base, u_long num)
{
    resource_entry_t *p, *q;

    for (p = map; ; p = p->next) {
	if (p->next == map)
	    return CS_BAD_BASE;
	if ((p->next->base == base) && (p->next->num == num))
	    break;
    }
    
    q = p->next;
    p->next = q->next;
    kfree_s(q, sizeof(resource_entry_t));
    return CS_SUCCESS;
} /* sub_interval */

/*======================================================================

    These find ranges of I/O ports or memory addresses that are not
    currently allocated by other devices.  In the future, the ranges
    should be controlled by AdjustResourceInfo.
    
======================================================================*/

static int find_io_region(resource_entry_t *io)
{
    int align;
    resource_entry_t *m;
    
    if (io->base != 0)
	return alloc_io_region(io);
    
    for (align = 1; align < io->num; align *= 2) ;
    for (m = io_db.next; m != &io_db; m = m->next) {
	for (io->base = m->base;
	     io->base+align <= m->base + m->num;
	     io->base += align)
	    if (alloc_io_region(io) == 0) return 0;
    }
    return -1;
} /* find_io_region */

static int find_mem_region(resource_entry_t *mem)
{
    int align;
    resource_entry_t *m;
    
    if (mem->base != 0)
	return alloc_mem_region(mem);
    
    for (align = 4096; align < mem->num; align *= 2) ;
    for (m = mem_db.next; m != &mem_db; m = m->next) {
	for (mem->base = m->base;
	     mem->base+align <= m->base+m->num;
	     mem->base += align)
	    if (alloc_mem_region(mem) == 0) return 0;
    }
    return -1;
} /* find_mem_region */

static int try_irq(irq_req_t *req, int irq, int specific)
{
    irq_info_t *info = &irq_table[irq];
    if (info->Attributes & RES_ALLOCATED) {
	switch (req->Attributes & IRQ_TYPE) {
	case IRQ_TYPE_EXCLUSIVE:
	    return CS_IN_USE;
	case IRQ_TYPE_TIME:
	    if ((info->Attributes & RES_IRQ_TYPE)
		!= RES_IRQ_TYPE_TIME)
		return CS_IN_USE;
	    if (req->Attributes & IRQ_FIRST_SHARED)
		return CS_BAD_ATTRIBUTE;
	    break;
	case IRQ_TYPE_DYNAMIC_SHARING:
	    if ((info->Attributes & RES_IRQ_TYPE)
		!= RES_IRQ_TYPE_DYNAMIC)
		return CS_IN_USE;
	    if (req->Attributes & IRQ_FIRST_SHARED)
		return CS_BAD_ATTRIBUTE;
	    break;
	}
    }
    else {
	if (((req->Attributes & IRQ_TYPE) != IRQ_TYPE_EXCLUSIVE)
	    && !(req->Attributes & IRQ_FIRST_SHARED))
	    return CS_BAD_ATTRIBUTE;
	if ((info->Attributes & RES_RESERVED) && !specific)
	    return CS_IN_USE;
	if (request_irq(irq, NULL) == -EBUSY)
	    return CS_IN_USE;
    }
    return 0;
} /* try_irq */

/*======================================================================

    Low-level functions to read and write attribute memory.  I think
    the write routine is only useful for writing one-byte registers.
    
======================================================================*/

static void read_attr_mem(socket_info_t *s, long addr, long len,
			  u_char *ptr)
{
    struct pcmcia_mem_map *mem = &s->attr_mem;
    u_char *sys = (u_char *)(mem->sys_start + (addr & 0x0fff));

    mem->card_start = addr & ~0x0fff;
    for (; len > 0; mem->card_start += 0x1000, sys -= 0x1000) {
	s->ctrl->ioctl(s->ctrl, PCMCIA_SET_MEM_MAP, mem);
	for ( ; len > 0; len--, ptr++, sys += 2) {
	    if (sys > (u_char *)mem->sys_stop) break;
	    *ptr = *sys;
	}
    }
} /* read_attr_mem */

static void write_attr_mem(socket_info_t *s, long addr, long len,
			   u_char *ptr)
{
    struct pcmcia_mem_map *mem = &s->attr_mem;
    u_char *sys = (u_char *)(mem->sys_start + (addr & 0x0fff));

    mem->card_start = addr & ~0x0fff;
    for (; len > 0; mem->card_start += 0x1000, sys -= 0x1000) {
	s->ctrl->ioctl(s->ctrl, PCMCIA_SET_MEM_MAP, mem);
	for ( ; len > 0; len--, ptr++, sys += 2) {
	    if (sys > (u_char *)mem->sys_stop) break;
	    *sys = *ptr;
	}
    }
} /* write_attr_mem */

static int setup_attr_mem(socket_info_t *s)
{
    if (s->attr_mem_alloc.base == 0) {
	if (find_mem_region(&s->attr_mem_alloc) != 0)
	    return CS_OUT_OF_RESOURCE;
	s->attr_mem.sys_start = s->attr_mem_alloc.base;
	s->attr_mem.sys_stop = s->attr_mem_alloc.base + 0x0fff;
    }
    return 0;
} /* setup_attr_mem */

/*======================================================================

    Low-level PCMCIA interface drivers need to register with Card
    Services using these calls.
    
======================================================================*/

static void setup_socket(u_long i);
static void shutdown_socket(u_long i);
static void unreset_socket(u_long i);
static void cs_interrupt(int irq);

int register_pcmcia(pcmcia_ctrl_t *ctrl)
{
    int status, irq;
    socket_info_t *s;
    irq_info_t *info;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk ("register_pcmcia(0x%p)\n", ctrl);
#endif
    if (sockets == MAX_SOCK) return -ENODEV;

    irq = ctrl->csc_irq;
    info = &irq_table[irq];
    if (info->Attributes & RES_ALLOCATED) {
	if (info->Attributes & RES_IRQ_CSC)
	    info->dyn_share++;
	else
	    return -EINVAL;
    }
    else {
	info->dyn_share = 1;
	info->Attributes |=
	    RES_ALLOCATED | RES_IRQ_CSC | RES_IRQ_TYPE_DYNAMIC;
	request_irq(irq, &cs_interrupt);
    }
    
#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif
    
    s = &socket_table[sockets];
    s->ctrl = ctrl;
    
    s->clients = NULL;
    s->setup.data = sockets;
    s->setup.function = &setup_socket;
    s->shutdown.data = sockets;
    s->shutdown.function = &shutdown_socket;
    s->unreset.data = sockets;
    s->unreset.function = &unreset_socket;
    s->attr_mem_alloc.base = 0;
    s->attr_mem_alloc.num = 0x1000;
    s->attr_mem.flags = MAP_ATTRIB | MAP_ACTIVE;
    s->attr_mem.map = 0;
    s->attr_mem.extra_ws = 0;
    sockets++;
    ctrl->ioctl(ctrl, PCMCIA_CSC_ON, &ctrl->csc_irq);
    ctrl->ioctl(ctrl, PCMCIA_GET_STATUS, &status);
    if (status & PC_DETECT)
	setup_socket(sockets-1);
    else
	s->state = 0;

    return 0;
} /* register_pcmcia */

/*====================================================================*/

void unregister_pcmcia(pcmcia_ctrl_t *ctrl)
{
    int i, irq;
    irq_info_t *info;
    
    for (i = 0; i < sockets; i++)
	if (socket_table[i].ctrl == ctrl) break;
    if (i == sockets)
	return;

    ctrl->ioctl(ctrl, PCMCIA_CSC_OFF, NULL);
    irq = ctrl->csc_irq;
    info = &irq_table[irq];
    info->dyn_share--;
    if (info->dyn_share == 0) {
	info->Attributes &= ~RES_RESERVED;
	free_irq(irq);
    }
    
    socket_table[i].ctrl = NULL;
    free_mem_region(&socket_table[i].attr_mem_alloc);
    
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
} /* unregister_pcmcia */

/*======================================================================

    Shutdown_Socket() and setup_socket() are scheduled using add_timer
    calls by the main event handler when card insertion and removal
    events are received.  Shutdown_Socket() unconfigures a socket and
    turns off socket power.  Setup_socket() turns on socket power
    and sends insertion events to valid clients.

======================================================================*/

static void send_event(socket_t s, long mask,
		       event_t event, int priority);

void shutdown_socket(u_long i)
{
    int val;
    socket_info_t *s = &socket_table[i];
    
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("cs: shutdown_socket(%ld)\n", i);
#endif
    
    s->state &= ~SOCKET_SHUTDOWN_PENDING;
    s->ctrl->ioctl(s->ctrl, PCMCIA_RESET, NULL);
    val = PC_PWR_OFF;
    s->ctrl->ioctl(s->ctrl, PCMCIA_SET_FLAGS, &val);
    s->state &= ~(SOCKET_CONFIG_LOCKED|SOCKET_IRQ_REQ|SOCKET_IO_REQ);
    s->irq.AssignedIRQ = 0;
    s->io.NumPorts1 = s->io.NumPorts2 = 0;
} /* shutdown_socket */

void setup_socket(u_long i)
{
    int val;
    socket_info_t *s = &socket_table[i];

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("cs: setup_socket(%ld)\n", i);
#endif

    s->state |= SOCKET_PRESENT;
    s->state &= ~SOCKET_SETUP_PENDING;
    val = PC_VCC_5V;
    s->ctrl->ioctl(s->ctrl, PCMCIA_SET_FLAGS, &val);
    send_event(i, EVENT_CARD_DETECT, CS_EVENT_CARD_INSERTION,
	       CS_EVENT_PRI_LOW);
} /* setup_socket */

void unreset_socket(u_long i)
{
    socket_info_t *s = &socket_table[i];
    int val;

    s->ctrl->ioctl(s->ctrl, PCMCIA_GET_FLAGS, &val);
    if (val | PC_RESET) {
	val &= ~PC_RESET;
	s->ctrl->ioctl(s->ctrl, PCMCIA_SET_FLAGS, &val);
    }
    
    s->ctrl->ioctl(s->ctrl, PCMCIA_GET_STATUS, &val);
    if (val & PC_READY) {
#ifdef PCMCIA_DEBUG
	if (pc_debug)
	    printk("cs: reset done on socket %ld\n", i);
#endif
	s->state &= ~SOCKET_RESET_PENDING;
	send_event(i, 0, CS_EVENT_CARD_RESET, CS_EVENT_PRI_LOW);
	s->reset_handle->event_callback_args.info = NULL;
	EVENT(s->reset_handle, CS_EVENT_RESET_COMPLETE, CS_EVENT_PRI_LOW);
    }
    else {
#ifdef PCMCIA_DEBUG
	if (pc_debug)
	    printk("cs: socket %ld not ready yet\n", i);
#endif
	if (s->unreset_timeout > 10)
	    printk("cs: socket %ld timed out\n", i);
	else {
	    s->unreset_timeout++;
	    s->unreset.expires = 20;
	    add_timer(&s->unreset);
	}
    }
}

/*======================================================================

    The central event handler.  Send_event() sends an event to all
    valid clients.  Parse_events() interprets the event bits from
    a card status change report.  And cs_interrupt() is the actual
    card status interrupt handler.
    
======================================================================*/

static void send_event(socket_t s, long mask,
		       event_t event, int priority)
{
    client_t *client = socket_table[s].clients;
#ifdef PCMCIA_DEBUG
    if (pc_debug > 1)
	printk("cs: send_event(sock %d, mask %#lx, event %d, pri %d)\n",
	       s, mask, event, priority);
#endif
    for (; client; client = client->next) { 
	if (client->state & (CLIENT_UNBOUND|CLIENT_STALE))
	    continue;
	if ((mask == 0) || (client->EventMask & mask))
	    EVENT(client, event, priority);
    }
} /* send_event */

static void parse_events(socket_t s, int events)
{
    socket_info_t *si = &socket_table[s];
    if (events & PC_DETECT) {
	client_t *client;
	int status;
	if ((si->state & SOCKET_PRESENT) &&
	    !(si->state & SOCKET_SHUTDOWN_PENDING)) {
	    si->state |= SOCKET_SHUTDOWN_PENDING;
	    send_event(s, EVENT_CARD_DETECT, CS_EVENT_CARD_REMOVAL,
		       CS_EVENT_PRI_HIGH);
	    for (client = si->clients; client; client = client->next)
		if (!(client->Attributes & INFO_DRIVER_SERVICES))
		    client->state |= CLIENT_STALE;
	    if (si->state & SOCKET_SETUP_PENDING) {
		printk("cs: flushing pending setup\n");
		del_timer(&si->setup);
		si->state &= ~SOCKET_SETUP_PENDING;
	    }
	    si->shutdown.expires = 1;
	    add_timer(&si->shutdown);
	}
	si->state &= ~SOCKET_PRESENT;
	si->ctrl->ioctl(si->ctrl, PCMCIA_GET_STATUS, &status);
	if (status & PC_DETECT) {
	    if (!(si->state & SOCKET_SETUP_PENDING)) {
		si->state |= SOCKET_SETUP_PENDING;
		si->setup.expires = 10;
		add_timer(&si->setup);
	    }
	}
    }
    if (events & PC_BATDEAD)
	send_event(s, EVENT_BATTERY_DEAD, CS_EVENT_BATTERY_DEAD,
		   CS_EVENT_PRI_LOW);
    if (events & PC_BATWARN)
	send_event(s, EVENT_BATTERY_LOW, CS_EVENT_BATTERY_LOW,
		   CS_EVENT_PRI_LOW);
    if (events & PC_READY) {
	if (si->state & SOCKET_RESET_PENDING)
	    printk("cs: ready change during reset on socket %d\n", s);
	else
	    send_event(s, EVENT_READY_CHANGE, CS_EVENT_READY_CHANGE,
		       CS_EVENT_PRI_LOW);
    }
} /* parse_events */

static void cs_interrupt(int irq)
{
    socket_t i;
    socket_info_t *s;
    int ret, hit;
    
    cli();
#ifdef PCMCIA_DEBUG
    if (pc_debug > 1)
	printk("cs: cs_interrupt()\n");
#endif
    do {
	hit = 0;
	for (i = 0, s = socket_table; i < sockets; i++, s++) {
	    if (s->ctrl == NULL)
		continue;
	    s->ctrl->ioctl(s->ctrl, PCMCIA_CSC_POLL, &ret);
#ifdef PCMCIA_DEBUG
	    if (pc_debug > 1)
		printk("cs: csc_poll(%d) = %#x\n", i, ret);
#endif
	    if (ret == 0) continue;
	    hit++;
	    parse_events(i, ret);
	}
    } while (hit > 0);
} /* cs_interrupt */

/*======================================================================

    Bind_device() associates a device driver with a particular socket.
    It is normally called by Driver Services after it has identified
    a newly inserted card.  An instance of that driver will then be
    eligible to register as a client of this socket.
    
======================================================================*/

int bind_device(bind_req_t *req)
{
    client_t *client;
    socket_info_t *s;
    u_long flags;

    if (CHECK_SOCKET(req->Socket))
	return CS_BAD_SOCKET;
    s = SOCKET(req);

    save_flags(flags);
    cli();
#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif
    client = (client_t *)kmalloc(sizeof(client_t), GFP_KERNEL);
    memset(client, '\0', sizeof(client_t));
    client->client_magic = CLIENT_MAGIC;
    client->dev_info = req->dev_info;
    client->Socket = req->Socket;
    client->state = CLIENT_UNBOUND;
    client->next = s->clients;
    s->clients = client;
    s->ctrl->open(s->ctrl);
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("cs: bind_device(): client 0x%p, sock %d, dev %s\n",
	       client, client->Socket, (char *)client->dev_info);
#endif
    restore_flags(flags);
    return CS_SUCCESS;
} /* bind_device */

/*======================================================================

    Access_configuration_register() reads and writes configuration
    registers in attribute memory.  Memory window 0 is reserved for
    this and the tuple reading services.
    
======================================================================*/

static int access_configuration_register(client_handle_t handle,
					 conf_reg_t *reg)
{
    socket_info_t *s;
    long addr;
    u_char val;
    
    if (CHECK_HANDLE(handle))
	return CS_BAD_HANDLE;
    if (!(handle->state & CLIENT_CONFIG_LOCKED))
	return CS_BAD_HANDLE;

    s = SOCKET(handle);
    if (setup_attr_mem(s) != 0)
	return CS_OUT_OF_RESOURCE;
    addr = (long)s->ConfigBase + reg->Offset;
    
    switch (reg->Action) {
    case CS_READ:
	read_attr_mem(s, addr, 1, &val);
	reg->Value = val;
	break;
    case CS_WRITE:
	val = reg->Value;
	write_attr_mem(s, addr, 1, &val);
	break;
    default:
	return CS_BAD_ARGS;
	break;
    }
    return CS_SUCCESS;
} /* access_configuration_register */

/*======================================================================

    We use standard Linux resource management for ports, address
    windows, and interrupts.  But we keep an internal list of
    things to avoid, for bogus drivers that don't allocate the
    stuff they use.
    
======================================================================*/

static int adjust_memory(adjust_t *adj)
{
    u_long base, num;

    base = (u_long)adj->resource.memory.Base;
    num = adj->resource.memory.Size;
    if ((num == 0) || (base+num-1 < base))
	return CS_BAD_SIZE;

    switch (adj->Action) {
    case ADD_MANAGED_RESOURCE:
	return add_interval(&mem_db, base, num);
	break;
    case REMOVE_MANAGED_RESOURCE:
	return sub_interval(&mem_db, base, num);
	break;
    default:
	return CS_UNSUPPORTED_FUNCTION;
	break;
    }
    
    return CS_SUCCESS;
} /* adjust_mem */

/*====================================================================*/

static int adjust_io(adjust_t *adj)
{
    int base, num;

    base = adj->resource.io.BasePort;
    num = adj->resource.io.NumPorts;
    if ((base < 0) || (base > 0xffff))
	return CS_BAD_BASE;
    if ((num <= 0) || (base+num > 0x10000) || (base+num <= base))
	return CS_BAD_SIZE;

    switch (adj->Action) {
    case ADD_MANAGED_RESOURCE:
	return add_interval(&io_db, base, num);
	break;
    case REMOVE_MANAGED_RESOURCE:
	return sub_interval(&io_db, base, num);
	break;
    default:
	return CS_UNSUPPORTED_FUNCTION;
	break;
    }
    
    return 0;
} /* adjust_io */

/*====================================================================*/

static int adjust_irq(adjust_t *adj)
{
    int irq;
    irq_info_t *info;
    
    irq = adj->resource.irq.IRQ;
    if ((irq < 0) || (irq > 15))
	return CS_BAD_IRQ;
    info = &irq_table[irq];
    
    switch (adj->Action) {
    case ADD_MANAGED_RESOURCE:
	if (info->Attributes & RES_REMOVED)
	    info->Attributes &= ~(RES_REMOVED|RES_ALLOCATED);
	else
	    if (adj->Attributes & RES_ALLOCATED)
		return CS_IN_USE;
	if (adj->Attributes & RES_RESERVED)
	    info->Attributes |= RES_RESERVED;
	else
	    info->Attributes &= ~RES_RESERVED;
	break;
    case REMOVE_MANAGED_RESOURCE:
	if (info->Attributes & RES_REMOVED)
	    return 0;
	if (info->Attributes & RES_ALLOCATED)
	    return CS_IN_USE;
	info->Attributes |= RES_ALLOCATED|RES_REMOVED;
	info->Attributes &= ~RES_RESERVED;
	break;
    default:
	return CS_UNSUPPORTED_FUNCTION;
	break;
    }
    
    return 0;
} /* adjust_irq */

/*====================================================================*/

static int adjust_resource_info(client_handle_t handle, adjust_t *adj)
{
    if (CHECK_HANDLE(handle))
	return CS_BAD_HANDLE;
    
    switch (adj->Resource) {
    case RES_MEMORY_RANGE:
	return adjust_memory(adj);
	break;
    case RES_IO_RANGE:
	return adjust_io(adj);
	break;
    case RES_IRQ:
	return adjust_irq(adj);
	break;
    }
    return CS_UNSUPPORTED_FUNCTION;
} /* adjust_resource_info */

/*====================================================================*/

static int deregister_client(client_handle_t handle)
{
    client_t **client;
    socket_info_t *s;
    u_long flags;
    
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("cs: deregister_client(%p)\n", handle);
#endif
    if (CHECK_HANDLE(handle))
	return CS_BAD_HANDLE;
    if (handle->state &
	(CLIENT_IRQ_REQ|CLIENT_IO_REQ|CLIENT_CONFIG_LOCKED))
	return CS_IN_USE;

    save_flags(flags);
    cli();
    s = SOCKET(handle);
    s->ctrl->close(s->ctrl);
    client = &s->clients;
    while ((*client) && ((*client) != handle))
	client = &(*client)->next;
    if (*client == NULL)
	return CS_BAD_HANDLE;
    *client = handle->next;
    kfree(handle);
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
    restore_flags(flags);
    return CS_SUCCESS;
} /* deregister_client */

/*====================================================================*/

static int get_configuration_info(client_handle_t handle,
				  config_t *config)
{
    int i;
    socket_info_t *s;
    
    if (CHECK_HANDLE(handle))
	return CS_BAD_HANDLE;
    i = handle->Socket; s = &socket_table[i];
    if (!(s->state & SOCKET_PRESENT))
	return CS_NO_CARD;
    if (!(s->state & SOCKET_CONFIG_LOCKED))
	return CS_CONFIGURATION_LOCKED;

    /* !!! This is a hack !!! */
    memcpy(&config->Attributes, &s->Attributes, 56);
    config->IRQAttributes = s->irq.Attributes;
    config->AssignedIRQ = s->irq.AssignedIRQ;
    config->BasePort1 = s->io.BasePort1;
    config->NumPorts1 = s->io.NumPorts1;
    config->Attributes1 = s->io.Attributes1;
    config->BasePort2 = s->io.BasePort2;
    config->NumPorts2 = s->io.NumPorts2;
    config->Attributes2 = s->io.Attributes2;
    config->IOAddrLines = s->io.IOAddrLines;
    
    return CS_SUCCESS;
} /* get_configuration_info */

/*======================================================================

    The Revision and CSLevel fields are sort-of just made up.
    
======================================================================*/

static int get_card_services_info(servinfo_t *info)
{
    info->InfoLen = sizeof(servinfo_t);
    info->Signature[0] = 'C';
    info->Signature[1] = 'S';
    info->Count = sockets;
    info->Revision = 0x0001;
    info->CSLevel = 0x0210;
    info->VendorString = (char *)version;
    return CS_SUCCESS;
} /* get_card_services_info */

/*======================================================================

    Note that get_first_client() *does* recognize the Socket field
    in the request structure.
    
======================================================================*/

static int get_first_client(client_handle_t *handle, client_req_t *req)
{
    socket_t s;
    if (req->Attributes & CLIENT_THIS_SOCKET)
	s = req->Socket;
    else
	s = 0;
    if (CHECK_SOCKET(req->Socket))
	return CS_BAD_SOCKET;
    if (socket_table[s].clients == NULL)
	return CS_NO_MORE_ITEMS;
    *handle = socket_table[s].clients;
    return CS_SUCCESS;
} /* get_first_client */

/*====================================================================*/

static int get_next_client(client_handle_t *handle, client_req_t *req)
{
    socket_info_t *s;
    if ((handle == NULL) || CHECK_HANDLE(*handle))
	return CS_BAD_HANDLE;
    if ((*handle)->next == NULL) {
	if (req->Attributes & CLIENT_THIS_SOCKET)
	    return CS_NO_MORE_ITEMS;
	s = SOCKET(*handle);
	if (s->clients == NULL)
	    return CS_NO_MORE_ITEMS;
	*handle = s->clients;
    }
    else
	*handle = (*handle)->next;
    return CS_SUCCESS;
} /* get_next_client */

/*======================================================================

    Read tuple codes from attribute memory.
    
======================================================================*/

static int get_next_tuple(client_handle_t handle, tuple_t *tuple);

static int get_first_tuple(client_handle_t handle, tuple_t *tuple)
{
    tuple->TupleLink = -1;
    return get_next_tuple(handle, tuple);
} /* get_first_tuple */

static int get_next_tuple(client_handle_t handle, tuple_t *tuple)
{
    socket_info_t *s;
    u_char link[2];
    long ofs = tuple->TupleLink;
    
    if (CHECK_HANDLE(handle))
	return CS_BAD_HANDLE;

    s = SOCKET(handle);
    if (setup_attr_mem(s) != 0)
	return CS_OUT_OF_RESOURCE;

    if (ofs < 0)
	ofs = 0;
    else {
	read_attr_mem(s, ofs, 2, link);
	ofs += (link[1]<<1) + 4;
    }
    
    if (tuple->DesiredTuple != RETURN_FIRST_TUPLE) {
	for (;;) {
	    read_attr_mem(s, ofs, 2, link);
	    if (link[0] == tuple->DesiredTuple)
		break;
	    if ((link[0] == CISTPL_NULL) || (link[0] == CISTPL_END))
		return CS_NO_MORE_ITEMS;
	    ofs += (link[1]<<1) + 4;
	}
    }

    tuple->TupleCode = link[0];
    tuple->TupleLink = ofs;
    return CS_SUCCESS;
} /* get_next_tuple */

/*====================================================================*/

static int get_status(client_handle_t handle, status_t *status)
{
    return CS_SUCCESS;
} /* get_status */

/*====================================================================*/

#define MIN(a,b) (((a)<(b)) ? (a) : (b))

static int get_tuple_data(client_handle_t handle, tuple_t *tuple)
{
    socket_info_t *s;
    u_char link[2];
    long len;
    
    if (CHECK_HANDLE(handle))
	return CS_BAD_HANDLE;

    s = SOCKET(handle);
    if (setup_attr_mem(s) != 0)
	return CS_OUT_OF_RESOURCE;

    read_attr_mem(s, tuple->TupleLink, 2, link);
    if (link[1] <= tuple->TupleOffset)
	return CS_NO_MORE_ITEMS;
    len = link[1] - tuple->TupleOffset;
    tuple->TupleDataLen = link[1];
    read_attr_mem(s, tuple->TupleLink+((tuple->TupleOffset+2)<<1),
		  MIN(len, tuple->TupleDataMax), tuple->TupleData);
    return CS_SUCCESS;
} /* get_tuple_data */

/*======================================================================

    Register_client() follows the Solaris strategy of ignoring the
    requested socket number; the dev_info_t handle is used to match
    the caller with a socket.  The driver must have already been
    bound to a socket with bind_device() -- in fact, bind_device()
    allocates the client structure that will be used.
    
======================================================================*/

static int register_client(client_handle_t *handle, client_reg_t *req)
{
    client_t *client;
    socket_t s;
    
    /* Look for unbound client with matching dev_info */
    client = NULL;
    for (s = 0; s < sockets; s++) {
	client = socket_table[s].clients;
	while (client != NULL) {
	    if ((client->dev_info == req->dev_info) &&
		(client->state & CLIENT_UNBOUND)) break;
	    client = client->next;
	}
	if (client != NULL) break;
    }
    
    if (client == NULL)
	return CS_OUT_OF_RESOURCE;
    *handle = client;
    client->state &= ~CLIENT_UNBOUND;
    client->Socket = s;
    client->Attributes = req->Attributes;
    client->EventMask = req->EventMask;
    client->event_handler = req->event_handler;
    client->event_callback_args = req->event_callback_args;
    client->event_callback_args.client_handle = client;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("cs: register_client(): client 0x%p, sock %d, dev %s\n",
	       client, client->Socket, (char *)client->dev_info);
#endif
    EVENT(client, CS_EVENT_REGISTRATION_COMPLETE, CS_EVENT_PRI_LOW);
    if (socket_table[s].state & SOCKET_PRESENT)
	EVENT(client, CS_EVENT_CARD_INSERTION, CS_EVENT_PRI_LOW);
    return CS_SUCCESS;
} /* register_client */

/*====================================================================*/

static int release_configuration(client_handle_t handle,
				 socket_t *Socket)
{
    if (CHECK_HANDLE(handle) ||
	!(handle->state & CLIENT_CONFIG_LOCKED))
	return CS_BAD_HANDLE;
    handle->state &= ~CLIENT_CONFIG_LOCKED;
    
    if (!(handle->state & CLIENT_STALE)) {
	socket_info_t *s = SOCKET(handle);
	s->ctrl->ioctl(s->ctrl, PCMCIA_RESET, NULL);
	s->state &= ~SOCKET_CONFIG_LOCKED;
    }
    
    return CS_SUCCESS;
} /* release_configuration */

/*======================================================================

    Release_io() releases the I/O ranges allocated by a client.  This
    may be invoked some time after a card ejection has already dumped
    the actual socket configuration, so if the client is "stale", we
    don't bother checking the port ranges against the current socket
    values.
    
======================================================================*/

static int release_io(client_handle_t handle, io_req_t *req)
{
    socket_info_t *s;
    
    if (CHECK_HANDLE(handle) || !(handle->state & CLIENT_IO_REQ))
	return CS_BAD_HANDLE;
    handle->state &= ~CLIENT_IO_REQ;
    s = SOCKET(handle);
    
    if (!(handle->state & CLIENT_STALE)) {
	if (s->state & SOCKET_CONFIG_LOCKED)
	    return CS_CONFIGURATION_LOCKED;
	if ((s->io.BasePort1 != req->BasePort1) ||
	    (s->io.NumPorts1 != req->NumPorts1) ||
	    (s->io.BasePort2 != req->BasePort2) ||
	    (s->io.NumPorts2 != req->NumPorts2))
	    return CS_BAD_ARGS;
	s->state &= ~SOCKET_IO_REQ;
    }

    free_io_region(&handle->io_1);
    if (handle->io_2.num)
	free_io_region(&handle->io_2);
    
    return CS_SUCCESS;
} /* release_io */

/*====================================================================*/

static int release_irq(client_handle_t handle, irq_req_t *req)
{
    irq_info_t *info;
    
    if (CHECK_HANDLE(handle) || !(handle->state & CLIENT_IRQ_REQ))
	return CS_BAD_HANDLE;
    handle->state &= ~CLIENT_IRQ_REQ;
    
    if (!(handle->state & CLIENT_STALE)) {
	socket_info_t *s = SOCKET(handle);
	if (s->state & SOCKET_CONFIG_LOCKED)
	    return CS_CONFIGURATION_LOCKED;
	if (s->irq.Attributes != req->Attributes)
	    return CS_BAD_ATTRIBUTE;
	if (s->irq.AssignedIRQ != req->AssignedIRQ)
	    return CS_BAD_IRQ;
	s->state &= ~SOCKET_IRQ_REQ;
    }
    
    info = &irq_table[req->AssignedIRQ];
    switch (req->Attributes & IRQ_TYPE) {
    case IRQ_TYPE_EXCLUSIVE:
	info->Attributes &= RES_RESERVED;
	break;
    case IRQ_TYPE_TIME:
	info->time_share--;
	if (info->time_share == 0)
	    info->Attributes &= RES_RESERVED;
	break;
    case IRQ_TYPE_DYNAMIC_SHARING:
	info->dyn_share--;
	if (info->dyn_share == 0)
	    info->Attributes &= RES_RESERVED;
	break;
    }
    
    return CS_SUCCESS;
} /* release_irq */

/*====================================================================*/

static int request_configuration(client_handle_t handle,
				 config_req_t *req)
{
    int i, val;
    socket_info_t *s;
    struct pcmcia_io_map iomap;
    
    if (CHECK_HANDLE(handle))
	return CS_BAD_HANDLE;
    i = handle->Socket; s = &socket_table[i];
    if (!(s->state & SOCKET_PRESENT))
	return CS_NO_CARD;
    if (s->state & SOCKET_CONFIG_LOCKED)
	return CS_CONFIGURATION_LOCKED;
    if (setup_attr_mem(s) != 0)
	return CS_OUT_OF_RESOURCE;

    /* Do power control */
    if (req->Vcc == 0)
	val = PC_PWR_OFF;
    else if (req->Vcc == 33)
	val = PC_VCC_3V;
    else if (req->Vcc == 50)
	val = PC_VCC_5V;
    else return CS_BAD_VCC;
    if (req->Vpp1 != req->Vpp2)
	return CS_BAD_VPP;
    if (req->Vpp1 == 50)
	val |= PC_VPP_5V;
    else if (req->Vpp1 == 120)
	val |= PC_VPP_12V;
    else if (req->Vpp1 != 0)
	return CS_BAD_VPP;
    s->ctrl->ioctl(s->ctrl, PCMCIA_SET_FLAGS, &val);
    s->Vcc = req->Vcc; s->Vpp1 = s->Vpp2 = req->Vpp1;
    
    /* Pick memory or I/O card */
    s->ctrl->ioctl(s->ctrl, PCMCIA_GET_FLAGS, &val);
    val &= ~PC_IOCARD;
    if (req->IntType & INT_MEMORY_AND_IO)
	val |= PC_IOCARD;
    s->ctrl->ioctl(s->ctrl, PCMCIA_SET_FLAGS, &val);
    s->IntType = req->IntType;
    
    /* Set up CIS configuration registers */
    s->ConfigBase = req->ConfigBase;
    s->Present = s->CardValues = req->Present;
    if (req->Present & PRESENT_OPTION) {
	s->Option = req->ConfigIndex & COR_CONFIG_MASK;
	if (s->state & SOCKET_IRQ_REQ)
	    if (!(s->irq.Attributes & IRQ_FORCED_PULSE))
		s->Option |= COR_LEVEL_REQ;
	write_attr_mem(s, (long)req->ConfigBase, 1, &s->Option);
	busy_loop(4);
    }
    if (req->Present & PRESENT_STATUS) {
	s->Status = req->Status;
	write_attr_mem(s, (long)req->ConfigBase+2, 1, &s->Status);
    }
    if (req->Present & PRESENT_PIN_REPLACE) {
	s->Pin = req->Pin;
	write_attr_mem(s, (long)req->ConfigBase+4, 1, &s->Pin);
    }
    if (req->Present & PRESENT_COPY) {
	s->Copy = req->Copy;
	write_attr_mem(s, (long)req->ConfigBase+6, 1, &s->Copy);
    }
    
    /* Configure I/O windows */
    if (s->state & SOCKET_IO_REQ) {
	if (s->io.NumPorts1) {
	    iomap.map = 0; iomap.extra_ws = 0;
	    iomap.flags = MAP_ACTIVE;
	    if (s->io.Attributes1 & IO_DATA_PATH_WIDTH)
		iomap.flags |= MAP_16BIT;
	    iomap.start = s->io.BasePort1;
	    iomap.stop = iomap.start + s->io.NumPorts1 - 1;
	    s->ctrl->ioctl(s->ctrl, PCMCIA_SET_IO_MAP, &iomap);
	}
	if (s->io.NumPorts2) {
	    iomap.map = 1; iomap.extra_ws = 0;
	    iomap.flags = MAP_ACTIVE;
	    if (s->io.Attributes2 & IO_DATA_PATH_WIDTH)
		iomap.flags |= MAP_16BIT;
	    iomap.start = s->io.BasePort2;
	    iomap.stop = iomap.start + s->io.NumPorts2 - 1;
	    s->ctrl->ioctl(s->ctrl, PCMCIA_SET_IO_MAP, &iomap);
	}
    }
    
    /* Configure I/O interrupt */
    if ((req->Attributes & CONF_ENABLE_IRQ) &&
	(s->state & SOCKET_IRQ_REQ)) {
	s->ctrl->ioctl(s->ctrl, PCMCIA_SET_IRQ, &s->irq.AssignedIRQ);
    }
    s->Attributes = req->Attributes;
    
    s->state |= SOCKET_CONFIG_LOCKED;
    handle->state |= CLIENT_CONFIG_LOCKED;
    return CS_SUCCESS;
} /* request_configuration */

/*======================================================================
  
    Request_io() reserves ranges of port addresses for a socket.
    I have not implemented range sharing or alias addressing.
    
======================================================================*/

static int request_io(client_handle_t handle, io_req_t *req)
{
    socket_info_t *s;
    
    if (CHECK_HANDLE(handle))
	return CS_BAD_HANDLE;
    s = SOCKET(handle);
    if (!(s->state & SOCKET_PRESENT))
	return CS_NO_CARD;
    if (s->state & SOCKET_IO_REQ)
	return CS_IN_USE;
    if (s->state & SOCKET_CONFIG_LOCKED)
	return CS_CONFIGURATION_LOCKED;
    if (req->Attributes1 & (IO_SHARED | IO_FORCE_ALIAS_ACCESS))
	return CS_BAD_ATTRIBUTE;
    if ((req->NumPorts2 > 0) &&
	(req->Attributes2 & (IO_SHARED | IO_FORCE_ALIAS_ACCESS)))
	return CS_BAD_ATTRIBUTE;

    handle->io_1.num = req->NumPorts1;
    handle->io_1.base = req->BasePort1;
    if (find_io_region(&handle->io_1))
	return CS_IN_USE;
    req->BasePort1 = handle->io_1.base;

    handle->io_2.num = req->NumPorts2;
    if (handle->io_2.num) {
	handle->io_2.base = req->BasePort2;
	if (find_io_region(&handle->io_2)) {
	    free_io_region(&handle->io_1);
	    return CS_IN_USE;
	}
	req->BasePort2 = handle->io_2.base;
    }

    s->io = *req;
    s->state |= SOCKET_IO_REQ;
    handle->state |= CLIENT_IO_REQ;
    return CS_SUCCESS;
} /* request_io */

/*======================================================================

    Request_irq() reserves an irq for this client.  I've tried to
    implement some support for shared interrupts, but for dynamic
    sharing, the handlers still need to chain themselves together.
    Also, since Linux only reserves irq's when they are actually
    hooked, we don't guarantee that an irq will still be available
    when the configuration is locked.  Now that I think about it,
    there might be a way to fix this using a dummy handler.
    
======================================================================*/

static int cs_request_irq(client_handle_t handle, irq_req_t *req)
{
    socket_info_t *s;
    int ret, irq;
    irq_info_t *info;
    
    if (CHECK_HANDLE(handle))
	return CS_BAD_HANDLE;
    s = SOCKET(handle);
    if (!(s->state & SOCKET_PRESENT))
	return CS_NO_CARD;
    if (s->state & SOCKET_IRQ_REQ)
	return CS_IN_USE;
    if (s->state & SOCKET_CONFIG_LOCKED)
	return CS_CONFIGURATION_LOCKED;
    
    ret = CS_BAD_ARGS;
    if (req->IRQInfo1 & IRQ_INFO2_VALID) {
	for (irq = 0; irq < 16; irq++)
	    if ((req->IRQInfo2 >> irq) & 1) {
		ret = try_irq(req, irq, 0);
		if (ret == 0) break;
	    }
	if (irq == 16)
	    for (irq = 0; irq < 16; irq++)
		if ((req->IRQInfo2 >> irq) & 1) {
		    ret = try_irq(req, irq, 1);
		    if (ret == 0) break;
		}
    }
    else {
	irq = req->IRQInfo1 & IRQ_MASK;
	ret = try_irq(req, irq, 1);
    }
    if (ret < 0) return ret;

    info = &irq_table[irq];
    info->Attributes |= RES_ALLOCATED;
    switch (req->Attributes & IRQ_TYPE) {
    case IRQ_TYPE_TIME:
	info->Attributes |= RES_IRQ_TYPE_TIME;
	info->time_share++;
	break;
    case IRQ_TYPE_DYNAMIC_SHARING:
	info->Attributes |= RES_IRQ_TYPE_DYNAMIC;
	info->dyn_share++;
	break;
    }
    
    req->AssignedIRQ = irq;
    s->irq = *req;
    s->state |= SOCKET_IRQ_REQ;
    handle->state |= CLIENT_IRQ_REQ;
    return CS_SUCCESS;
} /* cs_request_irq */

/*======================================================================

    I'm not sure which "reset" function this is supposed to use,
    but for now, it uses the low-level interface's reset, not the
    CIS register.
    
======================================================================*/

static int reset_card(client_handle_t handle, client_req_t *req)
{
    int i, val, ret;
    socket_info_t *s;
    client_t *client;
    
    if (CHECK_HANDLE(handle))
	return CS_BAD_HANDLE;
    i = handle->Socket; s = &socket_table[i];
    if (!(s->state & SOCKET_PRESENT))
	return CS_NO_CARD;
    if (s->state & SOCKET_RESET_PENDING)
	return CS_IN_USE;
    s->state |= SOCKET_RESET_PENDING;

    ret = 0;
    for (client = s->clients; client; client = client->next) {
	if (client->state & (CLIENT_UNBOUND|CLIENT_STALE))
	    continue;
	ret = EVENT(client, CS_EVENT_RESET_REQUEST, CS_EVENT_PRI_LOW);
	if (ret != 0) break;
    }
    if (client) {
	s->state &= ~SOCKET_RESET_PENDING;
	handle->event_callback_args.info = (void *)ret;
	EVENT(handle, CS_EVENT_RESET_COMPLETE, CS_EVENT_PRI_LOW);
    }
    else {
#ifdef PCMCIA_DEBUG
	if (pc_debug)
	    printk("cs: resetting socket %d\n", i);
#endif
	send_event(i, 0, CS_EVENT_RESET_PHYSICAL, CS_EVENT_PRI_LOW);
	s->reset_handle = handle;
	s->ctrl->ioctl(s->ctrl, PCMCIA_GET_FLAGS, &val);
	val |= PC_RESET;
	s->ctrl->ioctl(s->ctrl, PCMCIA_SET_FLAGS, &val);
	s->unreset_timeout = 0;
	s->unreset.expires = 20;
	add_timer(&s->unreset);
    }
    return CS_SUCCESS;
} /* reset_card */

/*======================================================================

    Maybe this should send a CS_EVENT_CARD_INSERTION event if we
    haven't sent one to this client yet?
    
======================================================================*/

static int set_event_mask(client_handle_t handle, eventmask_t *mask)
{
    if (handle->Attributes & CONF_EVENT_MASK_VALID)
	return CS_BAD_SOCKET;
    handle->EventMask = mask->EventMask;
    return CS_SUCCESS;
} /* set_event_mask */

/*====================================================================*/

int CardServices(int func, void *a1, void *a2, void *a3, void *a4)
{

#ifdef PCMCIA_DEBUG
    if (pc_debug > 1)
	printk("cs: CardServices(%d, 0x%p, 0x%p)\n", func, a1, a2);
#endif
    switch (func) {
    case AccessConfigurationRegister:
	return access_configuration_register(a1, a2); break;
    case AdjustResourceInfo:
	return adjust_resource_info(a1, a2); break;
    case DeregisterClient:
	return deregister_client(a1); break;
    case GetFirstClient:
	return get_first_client(a1, a2); break;
    case GetCardServicesInfo:
	return get_card_services_info(a1); break;
    case GetConfigurationInfo:
	return get_configuration_info(a1, a2); break;
    case GetNextClient:
	return get_next_client(a1, a2); break;
    case GetFirstTuple:
	return get_first_tuple(a1, a2); break;
    case GetNextTuple:
	return get_next_tuple(a1, a2); break;
    case GetStatus:
	return get_status(a1, a2); break;
    case GetTupleData:
	return get_tuple_data(a1, a2); break;
    case ParseTuple:
	return parse_tuple(a1, a2, a3); break;
    case RegisterClient:
	return register_client(a1, a2); break;
    case ReleaseConfiguration:
	return release_configuration(a1, a2); break;
    case ReleaseIO:
	return release_io(a1, a2); break;
    case ReleaseIRQ:
	return release_irq(a1, a2); break;
    case RequestConfiguration:
	return request_configuration(a1, a2); break;
    case RequestIO:
	return request_io(a1, a2); break;
    case RequestIRQ:
	return cs_request_irq(a1, a2); break;
    case ResetCard:
	return reset_card(a1, a2); break;
    case SetEventMask:
	return set_event_mask(a1, a2); break;
    default:
	return CS_UNSUPPORTED_FUNCTION; break;
    };
    
} /* CardServices */

/*====================================================================*/

#ifdef MODULE
char kernel_version[] = UTS_RELEASE;

int init_module(void)
{
    cs_init(0,0);
    return 0;
}

void cleanup_module(void)
{
    printk("unloading PCMCIA Card Services\n");
}
#endif /* MODULE */
