/*======================================================================

    A non-shared-memory PCMCIA ethernet driver for linux

    This driver supports the D-Link DE-650 and Linksys EthernetCard
    cards, and should soon support the IBM Credit Card Adapter.

    Written by David Hinds, dhinds@allegro.stanford.edu

    The network driver code is based on Donald Becker's NE2000 code:
    
    Written 1992,1993 by Donald Becker.
    Copyright 1993 United States Government as represented by the
    Director, National Security Agency.  This software may be used and
    distributed according to the terms of the GNU Public License,
    incorporated herein by reference.
    Donald Becker may be reached at becker@cesdis1.gsfc.nasa.gov
    
======================================================================*/

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <asm/io.h>
#include <asm/system.h>

#include <linux/netdevice.h>
#include "../net/8390.h"

#ifdef MODULE
#include <linux/module.h>
#include "../../tools/version.h"
#endif

#include "cs_types.h"
#include "cs.h"
#include "cistpl.h"
#include "ds.h"

#define DE650_BASE	(dev->base_addr)
#define DE650_CMD	0x00
#define DE650_DATAPORT	0x10	/* NatSemi-defined port window offset. */
#define DE650_RESET	0x1f	/* Issue a read to reset, a write to clear. */

#define DE650_START_PG	0x40	/* First page of TX buffer */
#define DE650_STOP_PG	0x80	/* Last page +1 of RX ring */

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
static char *version =
    "de650_cs.c v$Revision: 1.15 $: $Date: 1994/06/24 01:26:47 $ (David Hinds)\n";
#endif

/*====================================================================*/

int de650_probe(struct device *dev);

static void de650_config(u_long arg);
static void de650_release(u_long arg);
static int de650_event(event_t event, int priority,
		       event_callback_args_t *args);

static int de650_open(struct device *dev);
static int de650_close(struct device *dev);

static void de650_reset_8390(struct device *dev);
static int de650_block_input(struct device *dev, int count,
			     char *buf, int ring_offset);
static void de650_block_output(struct device *dev, const int count,
			       const unsigned char *buf,
			       const int start_page);

static dev_info_t dev_info = "de650_cs";

static dev_link_t *de650_attach(void);
static void de650_detach(dev_link_t *);

static dev_link_t default_link = {
    "\0",
    0,
    { 0, 0, 0, 0, &de650_config }, { 0, 0, 0, 0, &de650_release },
    NULL,
    { 0, 0, 16, 0, 0, 16, IO_DATA_PATH_WIDTH, 5 },
    { 0, IRQ_TYPE_EXCLUSIVE, 0, IRQ_INFO2_VALID|IRQ_LEVEL_ID, 0x86bc },
    {
	0, CONF_ENABLE_IRQ, 50, 0, 0, INT_MEMORY_AND_IO,
	(caddr_t)0x0008, 0, 0, 0, 1, PRESENT_OPTION
    },
    NULL,
    NULL
};

static dev_link_t *dev_list = NULL;

/*======================================================================

    If we're linked into the kernel, this gets called from the net
    driver initialization code in Space.c.  It returns failure,
    since all de650 devices are dynamically allocated.
    
======================================================================*/

#ifndef MODULE
int de650_probe(struct device *dev)
{
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk(version);
#endif
    register_pcmcia_driver(&dev_info, &de650_attach, &de650_detach);
    return -1;
}
#endif

/*======================================================================

    We never need to do anything when a de650 device is "initialized"
    by the net software, because we only register already-found cards.
    
======================================================================*/

static int de650_init(struct device *dev)
{
    return 0;
}

/*======================================================================

    de650_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

======================================================================*/

static dev_link_t *de650_attach(void)
{
    client_reg_t client_reg;
    dev_link_t *link;
    struct device *dev;
    int ret;
    
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("de650_attach()\n");
#endif

#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif

    /* Create new ethernet device */
    link = kmalloc(sizeof(struct dev_link_t), GFP_KERNEL);
    *link = default_link;
    link->config.data = link->release.data = (u_long)link;
    dev = kmalloc(sizeof(struct device), GFP_KERNEL);
    memset(dev, 0, sizeof(struct device));
    dev->name = link->dev_name;
    dev->init = &de650_init;
    dev->open = &de650_open;
    dev->tbusy = 1;
    link->priv = dev;
    
    /* Register with Card Services */
    link->next = dev_list;
    dev_list = link;
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask = EVENT_CARD_DETECT;
    client_reg.event_handler = &de650_event;
    client_reg.Version = 0x0210;
    client_reg.Socket = 0;
    ret = CardServices(RegisterClient, &link->handle, &client_reg);
    if (ret != 0) {
	printk("de650_cs: RegisterClient failed, ret = %d\n", ret);
	de650_detach(link);
	return NULL;
    }
    
    return link;
} /* de650_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/

static void de650_detach(dev_link_t *link)
{
    dev_link_t **linkp;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("de650_detach(0x%p)\n", link);
#endif
    
    /* Locate device structure */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
	if (*linkp == link) break;
    if (*linkp == NULL)
	return;

    if (link->state & DEV_CONFIG) {
	printk("de650_cs: detach postponed, '%s' still locked\n",
	       link->dev_name);
	link->state |= DEV_STALE_LINK;
	return;
    }
    
    if (link->handle)
	CardServices(DeregisterClient, link->handle);
    
    /* Unlink device structure, free bits */
    *linkp = link->next;
    kfree_s(link->priv, sizeof(struct device));
    kfree_s(link, sizeof(struct dev_link_t));
    
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
    
} /* de650_detach */

/*======================================================================

    de650_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    ethernet device available to the system.
    
======================================================================*/

static void de650_config(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;
    client_handle_t handle;
    tuple_t tuple;
    struct device *dev;
    int i;
    char *id;
    
    handle = link->handle;
    dev = link->priv;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("de650_config(0x%p)\n", link);
#endif

    /* Get the hardware address from the CIS */
    tuple.TupleDataMax = 6;
    tuple.TupleData = dev->dev_addr;

    do {
	/* The DE-650 puts the address in the DEVICE tuple */
	tuple.DesiredTuple = CISTPL_DEVICE;
	i = CardServices(GetFirstTuple, handle, &tuple);
	if (i != CS_SUCCESS) break;
	tuple.TupleOffset = 0x1e;
	i = CardServices(GetTupleData, handle, &tuple);
	id = "D-Link DE-650";
	
	if (i != CS_SUCCESS) {
	    /* The Linksys card uses the CONFIG tuple */
	    tuple.DesiredTuple = CISTPL_CONFIG;
	    i = CardServices(GetFirstTuple, handle, &tuple);
	    if (i != CS_SUCCESS) break;
	    tuple.TupleOffset = 0x0e;
	    i = CardServices(GetTupleData, handle, &tuple);
	    if (i != CS_SUCCESS) break;
	    id = "Linksys E-Card";
	}
	if ((dev->dev_addr[0] != 0x00) || (dev->dev_addr[1] != 0x80) ||
	    (dev->dev_addr[2] != 0xC8))
	    i = -1;
    } while (0);

    if (i != CS_SUCCESS) {
	printk("de650_cs: unable to parse hardware address\n");
	link->state &= ~DEV_CONFIG_PENDING;
	return;
    }

    /* Configure card */
    link->state |= DEV_CONFIG;

#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif

    do {
	i = CardServices(RequestIO, link->handle, &link->io);
	if (i != CS_SUCCESS) {
	    printk("de650_cs: RequestIO failed, ret = %d\n", i);
	    break;
	}
	i = CardServices(RequestIRQ, link->handle, &link->irq);
	if (i != CS_SUCCESS) {
	    printk("de650_cs: RequestIRQ failed, ret = %d\n", i);
	    break;
	}
	i = CardServices(RequestConfiguration, link->handle, &link->conf);
	if (i != CS_SUCCESS) {
	    printk("de650_cs: RequestConfiguration failed, ret = %d\n", i);
	    break;
	}
	i = request_irq(link->irq.AssignedIRQ, &ei_interrupt);
	if (i != CS_SUCCESS) {
	    printk("de650_cs: unable to allocate irq %ld\n",
		   link->irq.AssignedIRQ);
	    break;
	}
	dev->irq = link->irq.AssignedIRQ;
	dev->base_addr = link->io.BasePort1;
	dev->tbusy = 0;
	i = register_netdev(dev);
	if (i != 0) {
	    printk("de650_cs: register_netdev() failed\n");
	    break;
	}
    } while (0);
    
    link->state &= ~DEV_CONFIG_PENDING;
    if (i != 0) {
	de650_release((u_long)link);
	return;
    }
    
    printk("%s: %s PCMCIA: ", dev->name, id);
    for (i = 0; i < ETHER_ADDR_LEN; i++)
	printk(" %02X", dev->dev_addr[i]);
    printk("\n");
    
} /* de650_config */

/*======================================================================

    After a card is removed, de650_release() will unregister the net
    device, and release the PCMCIA configuration.  If the device is
    still open, this will be postponed until it is closed.
    
======================================================================*/

static void de650_release(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;
    struct device *dev = link->priv;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("de650_release(0x%p)\n", link);
#endif
    
    if (dev->start) {
	printk("de650_cs: release postponed, '%s' still open\n",
	       link->dev_name);
	link->state |= DEV_STALE_CONFIG;
	return;
    }
    
    if (link->dev_name[0] != '\0')
	unregister_netdev(dev);
    CardServices(ReleaseConfiguration, link->handle);
    CardServices(ReleaseIO, link->handle, &link->io);
    CardServices(ReleaseIRQ, link->handle, &link->irq);
    if (dev->irq != 0) {
	free_irq(dev->irq);
	irq2dev_map[dev->irq] = NULL;
    }
    
    link->state &= ~DEV_CONFIG;
    if (link->state & DEV_STALE_LINK)
	de650_detach(link);
    
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
} /* de650_release */

/*======================================================================

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.  A CARD_REMOVAL event
    also sets some flags to discourage the net drivers from trying
    to talk to the card any more.
    
======================================================================*/

static int de650_event(event_t event, int priority,
		       event_callback_args_t *args)
{
    dev_link_t *link;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("de650_event()\n");
#endif
    
    /* Locate device structure */
    for (link = dev_list; link; link = link->next)
	if (link->handle == args->client_handle) break;
    if (link == NULL) {
	printk("de650_cs: bad event handle 0x%p!\n", args->client_handle);
	return 0;
    }
    
    switch (event) {
#ifdef PCMCIA_DEBUG
    case CS_EVENT_REGISTRATION_COMPLETE:
	if (pc_debug)
	    printk("de650_cs: registration complete\n");
	break;
#endif
    case CS_EVENT_CARD_REMOVAL:
	link->state &= ~DEV_PRESENT;
	if (link->state & DEV_CONFIG) {
	    ((struct device *)link->priv)->tbusy = 1;
	    link->release.expires = 5;
	    add_timer(&link->release);
	}
	break;
    case CS_EVENT_CARD_INSERTION:
	link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
	link->config.expires = 5;
	add_timer(&link->config);
	break;
    case CS_EVENT_RESET_REQUEST:
	return 1;
	break;
    }
    return 0;
} /* de650_event */

/*====================================================================*/

static int de650_open(struct device *dev)
{
    dev_link_t *link;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("de650_open('%s')\n", dev->name);
#endif
    
    for (link = dev_list; link; link = link->next)
	if (link->priv == dev) break;
    if ((link == NULL) || (link->state & DEV_STALE_CONFIG))
	return -ENODEV;
    
#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif
    
    ethdev_init(dev);
    ei_status.name = "D-Link / Linksys PCMCIA";
    ei_status.tx_start_page = DE650_START_PG;
    ei_status.rx_start_page = ei_status.tx_start_page + TX_PAGES;
    ei_status.stop_page = DE650_STOP_PG;
    ei_status.reset_8390 = &de650_reset_8390;
    ei_status.block_input = &de650_block_input;
    ei_status.block_output = &de650_block_output;
    dev->stop = &de650_close;
    return ei_open(dev);
} /* de650_open */

/*====================================================================*/

static int de650_close(struct device *dev)
{
    dev_link_t *link;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("de650_close('%s')\n", dev->name);
#endif
    
    for (link = dev_list; link; link = link->next)
	if (link->priv == dev) break;
    if (link == NULL)
	return -ENODEV;
    
    dev->start = 0;
    if (dev->priv != NULL)
	kfree_s(dev->priv, sizeof(struct ei_device));
    if (link->state & DEV_STALE_CONFIG)
	de650_release((u_long)link);
    
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
    
    return 0;
} /* de650_close */

/*======================================================================
  
    Hard reset the card.  This used to pause for the same period that
    a 8390 reset command required, but that shouldn't be necessary.
  
======================================================================*/
  
static void de650_reset_8390(struct device *dev)
{
    int tmp = inb_p(DE650_BASE + DE650_RESET);
    int reset_start_time = jiffies;
     
    if (ei_debug > 1) printk("resetting the 8390 t=%ld...", jiffies);
    ei_status.txing = 0;
     
    outb_p(tmp, DE650_BASE + DE650_RESET);
    /* This check _should_not_ be necessary, omit eventually. */
    while ((inb_p(DE650_BASE+EN0_ISR) & ENISR_RESET) == 0)
	if (jiffies - reset_start_time > 2) {
	    printk("%s: de650_reset_8390() did not complete.\n", dev->name);
	    break;
	}
} /* de650_reset_8390 */

/*======================================================================

    Block input and output, similar to the Crynwr packet driver.

======================================================================*/

static int de650_block_input(struct device *dev, int count,
			     char *buf, int ring_offset)
{
    int xfer_count = count;
    
    if (ei_debug > 4)
	if (count != 4) printk("%s: [bi=%d]\n", dev->name, count+4);
    if (ei_status.dmaing) {
	if (ei_debug > 0)
	    printk("%s: DMAing conflict in de650_block_input."
		   "[DMAstat:%1x][irqlock:%1x]\n",
		   dev->name, ei_status.dmaing, ei_status.irqlock);
	return 0;
    }
    ei_status.dmaing |= 0x01;
    outb_p(E8390_NODMA+E8390_PAGE0+E8390_START, DE650_BASE+DE650_CMD);
    outb_p(count & 0xff, DE650_BASE + EN0_RCNTLO);
    outb_p(count >> 8, DE650_BASE + EN0_RCNTHI);
    outb_p(ring_offset & 0xff, DE650_BASE + EN0_RSARLO);
    outb_p(ring_offset >> 8, DE650_BASE + EN0_RSARHI);
    outb_p(E8390_RREAD+E8390_START, DE650_BASE+DE650_CMD);
    if (ei_status.word16) {
	insw(DE650_BASE + DE650_DATAPORT,buf,count>>1);
	if (count & 0x01)
	    buf[count-1] = inb(DE650_BASE + DE650_DATAPORT), xfer_count++;
    } else {
	insb(DE650_BASE + DE650_DATAPORT, buf, count);
    }
    
    /* This was for the ALPHA version only, but enough people have
       encountering problems that it is still here. */
    if (ei_debug > 1) {		/* DMA termination address check... */
	int addr, tries = 20;
	do {
	    /* DON'T check for 'inb_p(EN0_ISR) & ENISR_RDC' here
	       -- it's broken! Check the "DMA" address instead. */
	    int high = inb_p(DE650_BASE + EN0_RSARHI);
	    int low = inb_p(DE650_BASE + EN0_RSARLO);
	    addr = (high << 8) + low;
	    if (ei_status.word16) addr -= 8; else addr -= 4;
	    if (((ring_offset + xfer_count) & 0xff) == (addr & 0xff))
		break;
	} while (--tries > 0);
	if (tries <= 0)
	    printk("%s: RX transfer address mismatch,"
		   "%#4.4x (expected) vs. %#4.4x (actual).\n",
		   dev->name, ring_offset + xfer_count, addr);
    }
    ei_status.dmaing &= ~0x01;
    return ring_offset + count;
} /* de650_block_input */
  
/*====================================================================*/

static void de650_block_output(struct device *dev, int count,
			       const unsigned char *buf,
			       const int start_page)
{
    int retries = 0;
    
    if (ei_debug > 4) printk("%s: [bo=%d]\n", dev->name, count);
    /* Round the count up for word writes.  Do we need to do this?
       What effect will an odd byte count have on the 8390?
       I should check someday. */
    if (ei_status.word16 && (count & 0x01))
	count++;
    if (ei_status.dmaing) {
	if (ei_debug > 0) 
	    printk("%s: DMAing conflict in de650_block_output."
		   "[DMAstat:%1x][irqlock:%1x]\n",
		   dev->name, ei_status.dmaing, ei_status.irqlock);
	  return;
    }
    ei_status.dmaing |= 0x02;
    /* We should already be in page 0, but to be safe... */
    outb_p(E8390_PAGE0+E8390_START+E8390_NODMA, DE650_BASE+DE650_CMD);
    
 retry:
    /* Now the normal output. */
    outb_p(count & 0xff, DE650_BASE + EN0_RCNTLO);
    outb_p(count >> 8,   DE650_BASE + EN0_RCNTHI);
    outb_p(0x00, DE650_BASE + EN0_RSARLO);
    outb_p(start_page, DE650_BASE + EN0_RSARHI);
    
    outb_p(E8390_RWRITE+E8390_START, DE650_BASE+DE650_CMD);
    if (ei_status.word16) {
	outsw(DE650_BASE + DE650_DATAPORT, buf, count>>1);
    } else {
	outsb(DE650_BASE + DE650_DATAPORT, buf, count);
    }
    
    /* This was for the ALPHA version only, but enough people have
       encountering problems that it is still here. */
    if (ei_debug > 1) {	/* DMA termination address check... */
	int addr, tries = 20;
	do {
	    /* DON'T check for 'inb_p(EN0_ISR) & ENISR_RDC' here
	       -- it's broken! Check the "DMA" address instead. */
	    int high = inb_p(DE650_BASE + EN0_RSARHI);
	    int low = inb_p(DE650_BASE + EN0_RSARLO);
	    addr = (high << 8) + low;
	    if ((start_page << 8) + count == addr)
		break;
	} while (--tries > 0);
	if (tries <= 0) {
	    printk("%s: Tx packet transfer address mismatch,"
		   "%#4.4x (expected) vs. %#4.4x (actual).\n",
		   dev->name, (start_page << 8) + count, addr);
	    if (retries++ == 0)
		goto retry;
	}
    }
    ei_status.dmaing &= ~0x02;
    return;
} /* de650_block_output */

/*====================================================================*/

#ifdef MODULE
char kernel_version[] = UTS_RELEASE;

int init_module(void)
{
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk(version);
#endif
    register_pcmcia_driver(&dev_info, &de650_attach, &de650_detach);
    return 0;
}

void cleanup_module(void)
{
    printk("de650_cs: unloading\n");
    unregister_pcmcia_driver(&dev_info);
}
#endif /* MODULE */
