/*======================================================================

    PCMCIA Driver Services

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/major.h>
#include <linux/ioport.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/fcntl.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/ioctl.h>

#include <asm/segment.h>

#ifdef MODULE
#include <linux/module.h>
#include "../../tools/version.h"
#endif

#include "cs_types.h"
#include "cs.h"
#include "cistpl.h"
#include "ds.h"

#ifdef PCMCIA_DEBUG
int ds_debug = PCMCIA_DEBUG;
static const char *version =
    "ds.c v$Revision: 1.14 $: $Date: 1994/06/22 23:20:18 $ (David Hinds)\n";
#endif

/*====================================================================*/

static void report_event(u_long i);

static int ds_event(event_t event, int priority,
		    event_callback_args_t *args);
static int ds_read(struct inode *inode, struct file *file,
		   char *buf, int count);
static int ds_select(struct inode *inode, struct file *file,
		     int sel_type, select_table *wait);
static int ds_ioctl(struct inode * inode, struct file * file,
		    u_int cmd, u_long arg);
static int ds_open(struct inode * inode, struct file * file);
static void ds_release(struct inode * inode, struct file * file);

typedef struct driver_info_t {
    dev_info_t *dev_info;
    dev_link_t *(*attach)(void);
    void (*detach)(dev_link_t *);
    struct driver_info_t *next;
} driver_info_t;

typedef struct socket_bind_t {
    driver_info_t *driver;
    dev_link_t *instance;
    struct socket_bind_t *next;
} socket_bind_t;

/* Socket state information */
typedef struct socket_info_t {
    client_handle_t handle;
    long state;
    long event;
    struct wait_queue *queue;
    struct timer_list notify;
    socket_bind_t *bind;
} socket_info_t;

#define SOCKET_PRESENT		0x01
#define SOCKET_BUSY		0x10
#define SOCKET_EVENT_PENDING	0x20

/*====================================================================*/

/* Device driver ID passed to Card Services */
static dev_info_t dev_info = "Driver Services";

/* Linked list of all registered device drivers */
static driver_info_t *root_driver = NULL;

static int sockets = 0;
static socket_info_t *socket_table = NULL;

static struct file_operations ds_fops = {
    NULL,		/* lseek */
    ds_read,		/* read */
    NULL,		/* write */
    NULL,		/* readdir */
    ds_select,		/* select */
    ds_ioctl,		/* ioctl */
    NULL,		/* mmap */
    ds_open,		/* open */
    ds_release,		/* release */
    NULL		/* fsync */
};

/*====================================================================*/

u_long ds_init(u_long kmem_start, u_long kmem_end)
{
    client_reg_t client_reg;
    servinfo_t serv;
    bind_req_t bind;
    socket_info_t *s;
    int i, ret;

    printk("PCMCIA Driver Services level 0.1\n");
#ifdef PCMCIA_DEBUG
    if (ds_debug)
	printk(version);
#endif
    ret = CardServices(GetCardServicesInfo, &serv);
    sockets = serv.Count;
    socket_table = kmalloc(sockets*sizeof(socket_info_t), GFP_KERNEL);
    for (i = 0, s = socket_table; i < sockets; i++, s++) {
	s->state = s->event = 0;
	s->queue = NULL;
	s->handle = NULL;
	s->notify.data = i;
	s->notify.function = &report_event;
	s->bind = NULL;
    }
    
    /* Set up hotline to Card Services */
    client_reg.dev_info = bind.dev_info = &dev_info;
    client_reg.Attributes = INFO_DRIVER_SERVICES;
    client_reg.EventMask = EVENT_CARD_DETECT;
    client_reg.Version = 0x0210;
    client_reg.event_handler = &ds_event;
    for (i = 0; i < sockets; i++) {
	client_reg.Socket = bind.Socket = i;
	ret = bind_device(&bind);
	if (ret != CS_SUCCESS) {
	    printk("ds: unable to bind to socket!\n");
	    break;
	}
	ret = CardServices(RegisterClient, &socket_table[i].handle,
			   &client_reg);
	if (ret != CS_SUCCESS) {
	    printk("ds: unable to register with Card Services!\n");
	    break;
	}
    }
    
    /* Set up character device for user mode clients */
    if (register_chrdev(PCMCIA_MAJOR, "ds", &ds_fops)) {
	printk("unable to get major %d for Driver Services\n",
	       PCMCIA_MAJOR);
    }
    
    return kmem_start;
} /* ds_init */

/*====================================================================*/

void ds_finish(void)
{
    int i;
    unregister_chrdev(PCMCIA_MAJOR, "ds");
    for (i = 0; i < sockets; i++)
	CardServices(DeregisterClient, socket_table[i].handle);
    sockets = 0;
    kfree(socket_table);
    
} /* ds_finish */

/*======================================================================

    Register_pcmcia_driver() and unregister_pcmcia_driver() are used
    tell Driver Services that a PCMCIA client driver is available to
    be bound to sockets.
    
======================================================================*/

int register_pcmcia_driver(dev_info_t *dev_info,
			   dev_link_t *(*attach)(void),
			   void (*detach)(dev_link_t *))
{
    driver_info_t *driver;
#ifdef PCMCIA_DEBUG
    if (ds_debug)
	printk("ds: register_pcmcia_driver('%s')\n", (char *)dev_info);
#endif
#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif
    driver = kmalloc(sizeof(driver_info_t), GFP_KERNEL);
    driver->next = root_driver;
    driver->dev_info = dev_info;
    driver->attach = attach;
    driver->detach = detach;
    root_driver = driver;
    return 0;
} /* register_pcmcia_driver */

/*====================================================================*/

int unregister_pcmcia_driver(dev_info_t *dev_info)
{
    driver_info_t *target, **driver = &root_driver;
#ifdef PCMIA_DEBUG
    if (ds_debug)
	printk("ds: unregister_pcmcia_driver('%s')\n", (char *)dev_info);
#endif
    while ((*driver) && ((*driver)->dev_info != dev_info))
	driver = &(*driver)->next;
    if (*driver == NULL)
	return -1;
    target = *driver;
    *driver = target->next;
    kfree(target);
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
    return 0;
} /* unregister_pcmcia_driver */

/*======================================================================

    report_event() signals card status changes to the the user mode
    device.
    
======================================================================*/

void report_event(u_long i)
{
    socket_info_t *s = &socket_table[i];
    
#ifdef PCMCIA_DEBUG
    if (ds_debug)
	printk("ds: report_event(%ld)\n", i);
#endif
    if (s->state & SOCKET_BUSY)
	wake_up_interruptible(&s->queue);
}

/*======================================================================

    The card status event handler.
    
======================================================================*/

static int ds_event(event_t event, int priority,
		    event_callback_args_t *args)
{
    socket_info_t *s;
    int i;

#ifdef PCMCIA_DEBUG
    if (ds_debug)
	printk("ds: ds_event(%d, %d, 0x%p)\n",
	       event, priority, args->client_handle);
#endif
    for (i = 0, s = socket_table; i < sockets; i++, s++)
	if (s->handle == args->client_handle) break;
    
    switch (event) {
#ifdef PCMCIA_DEBUG
    case CS_EVENT_REGISTRATION_COMPLETE:
	if (ds_debug)
	    printk("ds: registration_complete on socket %d\n", i);
	break;
#endif
    case CS_EVENT_CARD_INSERTION:
	s->state |= SOCKET_PRESENT;
	s->event |= DS_EVENT_CARD_INSERTION;
#ifdef PCMCIA_DEBUG
	if (ds_debug)
	    printk("ds: card insertion on socket %d\n", i);
#endif
	break;
    case CS_EVENT_CARD_REMOVAL:
	s->state &= ~SOCKET_PRESENT;
	s->event |= DS_EVENT_CARD_REMOVAL;
	s->event &= ~DS_EVENT_CARD_INSERTION;
	printk("ds: card ejection on socket %d\n", i);
	break;
    case CS_EVENT_RESET_COMPLETE:
	s->event |= DS_EVENT_RESET_COMPLETE;
#ifdef PCMCIA_DEBUG
	if (ds_debug)
	    printk("ds: reset complete on socket %d\n", i);
#endif
	break;
    }
    
    if (s->event && !(s->state & SOCKET_EVENT_PENDING)) {
	s->state |= SOCKET_EVENT_PENDING;
	s->notify.expires = 1;
	add_timer(&s->notify);
    }
    return 0;
} /* ds_event */

/*======================================================================

    bind_request() connects a socket to a particular client driver.
    It looks up the specified device ID in the list of registered
    drivers, binds it to the socket, and tries to create an instance
    of the device.  unbind_request() deletes a driver instance.
    
======================================================================*/

static int bind_request(int i, bind_info_t *bind_info)
{
    struct driver_info_t *driver;
    socket_bind_t *b;
    bind_req_t bind_req;
    socket_info_t *s = &socket_table[i];

    for (driver = root_driver; driver; driver = driver->next)
	if (strcmp((char *)driver->dev_info,
		   (char *)bind_info->dev_info) == 0)
	    break;
    if (driver == NULL)
	return -ENODEV;

    for (b = s->bind; b; b = b->next)
	if (driver == b->driver)
	    break;
    if (b != NULL) {
	bind_info->instance = b->instance;
	return -EBUSY;
    }

    bind_req.Socket = i;
    bind_req.dev_info = driver->dev_info;
    if (bind_device(&bind_req) != 0) {
	printk("ds: unable to bind '%s' to socket %d\n",
	       (char *)dev_info, i);
	return -ENODEV;
    }
    
    bind_info->instance = driver->attach();
    if (bind_info->instance == NULL) {
	printk("ds: unable to creat instance of '%s'!\n",
	       (char *)bind_info->dev_info);
	return -ENODEV;
    }

    /* Add binding to list for this socket */
    b = kmalloc(sizeof(socket_bind_t), GFP_KERNEL);
    b->driver = driver;
    b->instance = bind_info->instance;
    b->next = s->bind;
    s->bind = b;
    
    return 0;
} /* bind_request */

/*====================================================================*/

static int get_device_info(int i, bind_info_t *bind_info)
{
    socket_info_t *s = &socket_table[i];
    socket_bind_t *b;
    
    for (b = s->bind; b; b = b->next)
	if (strcmp((char *)b->driver->dev_info,
		   (char *)bind_info->dev_info) == 0)
	    break;
    if (b == NULL)
	return -ENODEV;

    if (b->instance != bind_info->instance)
	return -EINVAL;
    if (b->instance->state & DEV_CONFIG_PENDING)
	return -EAGAIN;
    strncpy(bind_info->name, b->instance->dev_name, 8);
    bind_info->name[7] = '\0';
    
    return 0;
} /* get_device_info */

/*====================================================================*/

static int unbind_request(int i, bind_info_t *bind_info)
{
    socket_info_t *s = &socket_table[i];
    socket_bind_t **b, *c;

    for (b = &s->bind; *b; b = &(*b)->next)
	if (strcmp((char *)(*b)->driver->dev_info,
		   (char *)bind_info->dev_info) == 0)
	    break;
    if (*b == NULL)
	return -ENODEV;
    
    c = *b;
    c->driver->detach(c->instance);
    *b = c->next;
    kfree_s(c, sizeof(socket_bind_t));
    
    return 0;
} /* unbind_request */

/*======================================================================

    The user-mode PCMCIA device interface

    I use a really simple scheme for reporting card status changes.
    When a status change happens, the SOCKET_EVENT_PENDING flag is
    set.  Reading from the socket device returns the current status,
    and resets the PENDING flag.
    
======================================================================*/

static int ds_open(struct inode *inode, struct file *file)
{
    socket_t i = MINOR(inode->i_rdev);
    socket_info_t *s;

#ifdef PCMCIA_DEBUG
    if (ds_debug)
	printk("ds_open(socket %d)\n", i);
#endif
    if ((i >= sockets) || (sockets == 0))
	return -ENODEV;
    s = &socket_table[i];
#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif
    /* Only allow one writer at a time */
    if ((file->f_flags & O_ACCMODE) != O_RDONLY) {
	if (s->state & SOCKET_BUSY) return -EBUSY;
	s->state |= SOCKET_BUSY;
	if (s->state & SOCKET_PRESENT) {
	    s->event |= DS_EVENT_CARD_INSERTION;
	    s->state |= SOCKET_EVENT_PENDING;
	}
	else {
	    s->event = 0;
	    s->state &= ~SOCKET_EVENT_PENDING;
	}
    }
    return 0;
} /* ds_open */

/*====================================================================*/

static void ds_release(struct inode *inode, struct file *file)
{
    socket_t i = MINOR(inode->i_rdev);
    socket_info_t *s;

#ifdef PCMCIA_DEBUG
    if (ds_debug)
	printk("ds_release(socket %d)\n", i);
#endif
    if ((i >= sockets) || (sockets == 0))
	return;
    s = &socket_table[i];
    if ((file->f_flags & O_ACCMODE) != O_RDONLY)
	s->state &= ~SOCKET_BUSY;
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
} /* ds_release */

/*====================================================================*/

static int ds_read(struct inode *inode, struct file *file,
		   char *buf, int count)
{
    socket_t i = MINOR(inode->i_rdev);
    socket_info_t *s;
    int val;

#ifdef PCMCIA_DEBUG
    if (ds_debug)
	printk("ds_read(socket %d)\n", i);
#endif
    
    if ((i >= sockets) || (sockets == 0))
	return -ENODEV;
    s = &socket_table[i];
    if (count < 4)
	return -EINVAL;
    if ((file->f_flags & O_ACCMODE) == O_RDONLY)
	return -EBADF;
    if (!(s->state & SOCKET_EVENT_PENDING)) {
	interruptible_sleep_on(&s->queue);
	if (current->signal & ~current->blocked)
	    return -EINTR;
    }
    val = s->event; s->event = 0;
    s->state &= ~SOCKET_EVENT_PENDING;
    put_fs_long(val, buf);
    return 4;
} /* ds_read */

/*====================================================================*/

static int ds_select(struct inode *inode, struct file *file,
		     int sel_type, select_table *wait)
{
    socket_t i = MINOR(inode->i_rdev);
    socket_info_t *s;

#ifdef PCMCIA_DEBUG
    if (ds_debug > 1)
	printk("ds_select(socket %d)\n", i);
#endif
    
    if ((i >= sockets) || (sockets == 0))
	return -ENODEV;
    s = &socket_table[i];
    if (sel_type != SEL_IN)
	return 0;
    if ((file->f_flags & O_ACCMODE) == O_RDONLY)
	return -EBADF;
    if (s->state & SOCKET_EVENT_PENDING)
	return 1;
    select_wait(&s->queue, wait);
    return 0;
} /* ds_select */

/*====================================================================*/

static int ds_ioctl(struct inode * inode, struct file * file,
		    u_int cmd, u_long arg)
{
    socket_t i = MINOR(inode->i_rdev);
    socket_info_t *s;
    u_int size;
    int ret, err;
    ds_ioctl_arg_t buf;

#ifdef PCMCIA_DEBUG
    if (ds_debug)
	printk("ds_ioctl(socket %d, %#x, %#lx)\n", i, cmd, arg);
#endif
    
    if ((i >= sockets) || (sockets == 0))
	return -ENODEV;
    s = &socket_table[i];
    
    size = (cmd & IOCSIZE_MASK) >> IOCSIZE_SHIFT;
    if (size > sizeof(ds_ioctl_arg_t)) return -EINVAL;
    if (cmd & IOC_IN) {
	/* check for *_SET_* ioctl calls */
	if (!(cmd & IOC_OUT) || (cmd == DS_BIND_REQUEST)) {
	    if (!suser())
		return -EPERM;
	}
	err = verify_area(VERIFY_READ, (char *)arg, size);
	if (err) {
#ifdef PCMCIA_DEBUG
	    if (ds_debug)
		printk("ds_ioctl(): verify_read = %d\n", err);
#endif
	    return err;
	}
    }
    if (cmd & IOC_OUT) {
	err = verify_area(VERIFY_WRITE, (char *)arg, size);
	if (err) {
#ifdef PCMCIA_DEBUG
	    if (ds_debug)
		printk("ds_ioctl(): verify_write = %d\n", err);
#endif
	    return err;
	}
    }
    
    err = ret = 0;
    
    if (cmd & IOC_IN) memcpy_fromfs((char *)&buf, (char *)arg, size);
    
    switch (cmd) {
    case DS_ADJUST_RESOURCE_INFO:
	ret = CardServices(AdjustResourceInfo, s->handle, &buf.adjust);
	break;
    case DS_GET_CARD_SERVICES_INFO:
	ret = CardServices(GetCardServicesInfo, &buf.servinfo);
	break;
    case DS_GET_CONFIGURATION_INFO:
	ret = CardServices(GetConfigurationInfo, s->handle, &buf.config);
	break;
    case DS_GET_FIRST_TUPLE:
	ret = CardServices(GetFirstTuple, s->handle, &buf.tuple);
	break;
    case DS_GET_NEXT_TUPLE:
	ret = CardServices(GetNextTuple, s->handle, &buf.tuple);
	break;
    case DS_GET_TUPLE_DATA:
	buf.tuple.TupleData = buf.tuple_parse.data;
	ret = CardServices(GetTupleData, s->handle, &buf.tuple);
	break;
    case DS_PARSE_TUPLE:
	buf.tuple.TupleData = buf.tuple_parse.data;
	ret = CardServices(ParseTuple, s->handle, &buf.tuple,
			   &buf.tuple_parse.parse);
	break;
    case DS_RESET_CARD:
	ret = CardServices(ResetCard, s->handle, &buf.client_req);
	break;
    case DS_BIND_REQUEST:
	err = bind_request(i, &buf.bind_info);
	break;
    case DS_GET_DEVICE_INFO:
	err = get_device_info(i, &buf.bind_info);
	break;
    case DS_UNBIND_REQUEST:
	err = unbind_request(i, &buf.bind_info);
	break;
    }
    
    if ((err == 0) && (ret != CS_SUCCESS)) {
#ifdef PCMCIA_DEBUG
	if (ds_debug)
	    printk("ds_ioctl: ret = %d\n", ret);
#endif
	switch (ret) {
	case CS_BAD_SOCKET: case CS_NO_CARD:
	    err = -ENODEV; break;
	case CS_BAD_ARGS: case CS_BAD_ATTRIBUTE: case CS_BAD_IRQ:
	    err = -EINVAL; break;
	case CS_IN_USE:
	    err = -EBUSY; break;
	case CS_OUT_OF_RESOURCE:
	    err = -ENOSPC; break;
	case CS_UNSUPPORTED_FUNCTION:
	    err = -ENOSYS; break;
	default:
	    err = -EIO; break;
	}
    }
    
    if (cmd & IOC_OUT) memcpy_tofs((char *)arg, (char *)&buf, size);
     
    return err;
} /* ds_ioctl */

/*====================================================================*/

#ifdef MODULE
char kernel_version[] = UTS_RELEASE;

int init_module(void)
{
    ds_init(0,0);
    return 0;
}

void cleanup_module(void)
{
    ds_finish();
}
#endif /* MODULE */
