/*======================================================================

    Device driver for Intel 82365 and compatible PCMCIA controllers

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <sys/types.h>
#include <linux/fcntl.h>
#include <linux/string.h>

#include <asm/io.h>
#include <asm/bitops.h>
#include <asm/segment.h>
#include <asm/system.h>

#ifdef USER_DRV
#include "user_drv.h"
#else
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/malloc.h>
#endif

#ifdef MODULE
#include <linux/module.h>
#include "../../tools/version.h"
#endif

#include <linux/pcmcia.h>
#include "i82365.h"
#include "pd67xx.h"

#ifdef PCMCIA_DEBUG
int pc_debug = PCMCIA_DEBUG;
#else
int pc_debug = 0;
#endif

static const char *version =
    "i82365.c v$Revision: 1.7 $: $Date: 1994/06/15 06:10:57 $ (David Hinds)\n";

/*====================================================================*/

static int i365_open(struct pcmcia_ctrl_t *ctrl);
static void i365_close(struct pcmcia_ctrl_t *ctrl);
static int i365_ioctl(struct pcmcia_ctrl_t *ctrl, u_int cmd, void *arg);

/* Default controller description */
static pcmcia_ctrl_t i365_ctrl =
    { "i82365sl", 0, 0, 0, &i365_open, &i365_close, &i365_ioctl, NULL };

/* For the pcmcia_ctrl_t flag field */
#define IS_PD67XX	0x01

/* Linked list of controller descriptions */
static pcmcia_ctrl_t *i365_list = NULL;

/* I/O port allocation info */
static resource_entry_t i365_ports = { I365_BASE, 4, NULL };

/*====================================================================*/

static u_char i365_get(u_short sock, u_short reg)
{
    u_short port = I365_PORT(sock);
    u_char val = I365_REG(sock, reg);
    cli();
    outb_p(val, port); val = inb_p(port+1);
    sti();
    return val;
}

static void i365_set(u_short sock, u_short reg, u_char data)
{
    u_short port = I365_PORT(sock);
    u_char val = I365_REG(sock, reg);
    cli();
    outb_p(val, port); outb_p(data, port+1);
    sti();
}

static void i365_bset(u_short sock, u_short reg, u_char mask)
{
    u_char d = i365_get(sock, reg);
    d |= mask;
    i365_set(sock, reg, d);
}

static void i365_bclr(u_short sock, u_short reg, u_char mask)
{
    u_char d = i365_get(sock, reg);
    d &= ~mask;
    i365_set(sock, reg, d);
}

static u_short i365_get_pair(u_short sock, u_short reg)
{
    u_short a, b;
    a = i365_get(sock, reg);
    b = i365_get(sock, reg+1);
    return (a + (b<<8));
}

static void i365_set_pair(u_short sock, u_short reg, u_short data)
{
    i365_set(sock, reg, data & 0xff);
    i365_set(sock, reg+1, data >> 8);
}

/*====================================================================*/

int i365_init(void)
{
    int irqlist[] = { 15, 14, 12, 11, 10, 9, 7, 5, 4, 3, 0 };
    int i, val, id, sock, *irqp;
    pcmcia_ctrl_t *ctrl;

    if (pc_debug)
	printk(version);

    printk("Intel PCIC probe: ");
    
    id = sock = 0;
    if (check_io_region(&i365_ports) == 0)
	for (; sock < 8; sock++) {
	    val = i365_get(sock, I365_IDENT);
	    if ((val != 0x82) && (val != 0x83)) break;
	       id = val & 7;
	}
    
    if (sock == 0) {
	printk("not found.\n");
	return -ENODEV;
    }

    alloc_io_region(&i365_ports);

    /* Check for Cirrus CL-PD67xx chips */
    i365_set(0, PD67_CHIP_INFO, 0);
    val = i365_get(0, PD67_CHIP_INFO);
    if ((val & PD67_INFO_CHIP_ID) == PD67_INFO_CHIP_ID) {
	val = i365_get(0, PD67_CHIP_INFO);
	if ((val & PD67_INFO_CHIP_ID) == 0) {
	    if (val & PD67_INFO_SLOTS)
		i365_ctrl.name = "Cirrus CL-PD672x";
	    else
		i365_ctrl.name = "Cirrus CL-PD6710";
	    id = 8 - ((val & PD67_INFO_REV) >> 2);
	    i365_ctrl.flags |= IS_PD67XX;
	}
    }

    /* Pick one irq for all sockets */
    for (irqp = irqlist; *irqp != 0; irqp++)
	if (request_irq(*irqp, NULL) != -EBUSY) break;
    
    printk("%s rev %d found, %d sockets, ", i365_ctrl.name, id, sock);
    if (*irqp != 0)
	printk("status change on IRQ %d\n", *irqp);
    else
	printk("no status change IRQ\n");

    for (i = 0; i < sock; i++) {
	ctrl = kmalloc(sizeof(pcmcia_ctrl_t), GFP_KERNEL);
	*ctrl = i365_ctrl;
	ctrl->sock = i;
	ctrl->csc_irq = *irqp;
	if (register_pcmcia(ctrl) != 0) {
	    printk("i82365: register_pcmcia() failed, sock %d\n", i);
	    kfree_s(ctrl, sizeof(pcmcia_ctrl_t));
	    break;
	}
	ctrl->next = i365_list;
	i365_list = ctrl;
    }

    if (i365_list == NULL) {
	free_io_region(&i365_ports);
	return -ENODEV;
    }
    else
	return 0;
    
} /* i365_init */
  
/*====================================================================*/

static void i365_finish(void)
{
    pcmcia_ctrl_t *c, *d;

    for (c = i365_list; c; c = d) {
	unregister_pcmcia(c);
	d = c->next;
	kfree_s(c, sizeof(pcmcia_ctrl_t));
    }
    free_io_region(&i365_ports);
} /* i365_finish */

/*====================================================================*/

static int i365_open(struct pcmcia_ctrl_t *ctrl)
{
#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif
    return 0;
}

static void i365_close(struct pcmcia_ctrl_t *ctrl)
{
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
}

/*====================================================================*/

static void busy_loop(u_long len)
{
    u_long timeout = jiffies + len;
    u_long flags;
    save_flags(flags);
    sti();
    while (timeout >= jiffies)
	;
    restore_flags(flags);
} /* busy_loop */

/*====================================================================*/

static int i365_get_status(u_short sock, u_int *value)
{
    u_char status;
    status = i365_get(sock, I365_STATUS);
    *value = ((status & I365_CS_DETECT) == I365_CS_DETECT)
	? PC_DETECT : 0;
    if (i365_get(sock, I365_INTCTL) & I365_PC_IOCARD)
	*value |= (status & I365_CS_STSCHG) ? 0 : PC_STSCHG;
    else {
	*value |= (status & I365_CS_BVD1) ? 0 : PC_BATDEAD;
	*value |= (status & I365_CS_BVD2) ? 0 : PC_BATWARN;
    }
    *value |= (status & I365_CS_WRPROT) ? PC_WRPROT : 0;
    *value |= (status & I365_CS_READY) ? PC_READY : 0;
    *value |= (status & I365_CS_POWERON) ? PC_POWERON : 0;
    return 0;
} /* i365_get_status */

/*====================================================================*/

static int i365_get_flags(u_short sock, u_int *value)
{
    u_char reg, vcc, vpp;
    reg = i365_get(sock, I365_POWER);
    *value  = (reg & I365_PWR_AUTO) ? PC_PWR_AUTO : 0;
    vcc = reg & I365_VCC_MASK; vpp = reg & I365_VPP_MASK;
    *value |= (vcc == I365_VCC_3V) ? PC_VCC_3V : 0;
    *value |= (vcc == I365_VCC_5V) ? PC_VCC_5V : 0;
    *value |= (vpp == I365_VPP_5V) ? PC_VPP_5V : 0;
    *value |= (vpp == I365_VPP_12V) ? PC_VPP_12V : 0;
    reg = i365_get(sock, I365_INTCTL);
    *value |= (reg & I365_PC_IOCARD) ? PC_IOCARD : 0;
    *value |= (reg & I365_PC_RESET) ? 0 : PC_RESET;
    return 0;
} /* i365_get_flags */

/*====================================================================*/

static int i365_set_flags(u_short sock, u_int *value)
{
    u_char reg = i365_get(sock, I365_POWER);
    reg &= ~(I365_PWR_AUTO | I365_VCC_MASK | I365_VPP_MASK);
    reg |= (*value & PC_PWR_AUTO) ? I365_PWR_AUTO : 0;
    reg |= (*value & PC_VCC_3V) ? I365_VCC_3V|I365_PWR_OUT : 0;
    reg |= (*value & PC_VCC_5V) ? I365_VCC_5V|I365_PWR_OUT : 0;
    reg |= (*value & PC_VPP_5V) ? I365_VPP_5V : 0;
    reg |= (*value & PC_VPP_12V) ? I365_VPP_12V : 0;
    i365_set(sock, I365_POWER, reg);
    busy_loop(2);
    reg = i365_get(sock, I365_INTCTL);
    reg &= ~(I365_PC_IOCARD | I365_PC_RESET);
    /* Note that the reset signal is inverted */
    reg |= (*value & PC_RESET) ? 0 : I365_PC_RESET;
    reg |= (*value & PC_IOCARD) ? I365_PC_IOCARD : 0;
    i365_set(sock, I365_INTCTL, reg);
    return 0;
} /* i365_set_flags */

/*====================================================================*/

static int i365_csc_on(u_short sock, int *irq)
{
    /* Turn on all card status change interrupts */
    i365_set(sock, I365_CSCINT, (*irq << 4) | I365_CSC_ANY);
    i365_bclr(sock, I365_INTCTL, I365_INTR_ENA);
    /* flush current card status changes */
    i365_get(sock, I365_CSC);
    return 0;
} /* i365_csc_on */

/*====================================================================*/

static int i365_csc_off(u_short sock, void *arg)
{
    i365_set(sock, I365_CSCINT, 0);
    return 0;
} /* i365_csc_off */

/*====================================================================*/

static int i365_csc_poll(u_short sock, int *ret)
{
    /* Anything to report? */
    int csc = i365_get(sock, I365_CSC);
    if (i365_get(sock, I365_INTCTL) & I365_PC_IOCARD)
	*ret  = (csc & I365_CSC_STSCHG) ? PC_STSCHG : 0;
    else {
	*ret  = (csc & I365_CSC_BVD1) ? PC_BATDEAD : 0;
	*ret |= (csc & I365_CSC_BVD2) ? PC_BATWARN : 0;
    }
    *ret |= (csc & I365_CSC_READY) ? PC_READY : 0;
    *ret |= (csc & I365_CSC_DETECT) ? PC_DETECT : 0;
    return 0;
} /* i365_csc_poll */

/*====================================================================*/

static int i365_get_irq(u_short sock, u_int *value)
{
    u_char ctl = i365_get(sock, I365_INTCTL);
    *value = ctl & I365_IRQ_MASK;
    return 0;
} /* i365_get_irq */

/*====================================================================*/

static int i365_set_irq(u_short sock, u_int *value)
{
    u_char ctl = i365_get(sock, I365_INTCTL);
    if (*value > 15) return -EINVAL;
    ctl = (ctl & ~I365_IRQ_MASK) | *value;
    i365_set(sock, I365_INTCTL, ctl);
    return 0;
} /* i365_set_irq */

/*====================================================================*/

static int i365_reset(u_short sock, void *arg)
{
    /* Turn off interrupts, select memory-only interface */
    i365_bclr(sock, I365_INTCTL, I365_PC_IOCARD | I365_IRQ_MASK);
    /* Turn off I/O windows */
    i365_bclr(sock, I365_ADDRWIN, I365_ENA_IO(0) | I365_ENA_IO(1));
    return 0;
} /* i365_reset */

/*====================================================================*/

static int i365_get_io_map(u_short sock, struct pcmcia_io_map *io)
{
    u_char map, ioctl, addr;
    map = io->map;
    if (map > 1) return -EINVAL;
    io->start = i365_get_pair(sock, I365_IO(map)+I365_W_START);
    io->stop = i365_get_pair(sock, I365_IO(map)+I365_W_STOP);
    ioctl = i365_get(sock, I365_IOCTL);
    addr = i365_get(sock, I365_ADDRWIN);
    io->extra_ws = (ioctl & I365_IOCTL_WAIT(map)) ? 1 : 0;
    io->flags  = (addr & I365_ENA_IO(map)) ? MAP_ACTIVE : 0;
    io->flags |= (ioctl & I365_IOCTL_0WS(map)) ? MAP_0WS : 0;
    io->flags |= (ioctl & I365_IOCTL_CS16(map)) ? MAP_CS16 : 0;
    io->flags |= (ioctl & I365_IOCTL_16BIT(map)) ? MAP_16BIT : 0;
    return 0;
} /* i365_get_io_map */

/*====================================================================*/

static int i365_set_io_map(u_short sock, struct pcmcia_io_map *io)
{
    u_char map, ioctl;
    map = io->map;
    if ((map > 1) || (io->start > 0xffff) || (io->stop > 0xffff) ||
	(io->stop < io->start)) return -EINVAL;
    /* Turn off the window before changing anything */
    if (i365_get(sock, I365_ADDRWIN) & I365_ENA_IO(map))
	i365_bclr(sock, I365_ADDRWIN, I365_ENA_IO(map));
    i365_set_pair(sock, I365_IO(map)+I365_W_START, io->start);
    i365_set_pair(sock, I365_IO(map)+I365_W_STOP, io->stop);
    ioctl = i365_get(sock, I365_IOCTL) & ~I365_IOCTL_MASK(map);
    if (io->extra_ws) ioctl |= I365_IOCTL_WAIT(map);
    if (io->flags & MAP_0WS) ioctl |= I365_IOCTL_0WS(map);
    if (io->flags & MAP_CS16) ioctl |= I365_IOCTL_CS16(map);
    if (io->flags & MAP_16BIT) ioctl |= I365_IOCTL_16BIT(map);
    i365_set(sock, I365_IOCTL, ioctl);
    /* Turn on the window if necessary */
    if (io->flags & MAP_ACTIVE)
	i365_bset(sock, I365_ADDRWIN, I365_ENA_IO(map));
    return 0;
} /* i365_set_io_map */

/*====================================================================*/

static int i365_get_mem_map(u_short sock, struct pcmcia_mem_map *mem)
{
    u_short base, i;
    u_char map, addr;
    
    map = mem->map;
    if (map > 4) return -EINVAL;
    addr = i365_get(sock, I365_ADDRWIN);
    mem->flags = (addr & I365_ENA_MEM(map)) ? MAP_ACTIVE : 0;
    base = I365_MEM(map);
    
    i = i365_get_pair(sock, base+I365_W_START);
    mem->flags |= (i & I365_MEM_16BIT) ? MAP_16BIT : 0;
    mem->flags |= (i & I365_MEM_0WS) ? MAP_0WS : 0;
    mem->sys_start = ((long)(i & 0x0fff) << 12);
    
    i = i365_get_pair(sock, base+I365_W_STOP);
    mem->extra_ws  = (i & I365_MEM_WS0) ? 1 : 0;
    mem->extra_ws += (i & I365_MEM_WS1) ? 2 : 0;
    mem->sys_stop = ((long)(i & 0x0fff) << 12) + 0x0fff;
    
    i = i365_get_pair(sock, base+I365_W_OFF);
    mem->flags |= (i & I365_MEM_WRPROT) ? MAP_WRPROT : 0;
    mem->flags |= (i & I365_MEM_REG) ? MAP_ATTRIB : 0;
    mem->card_start = ((long)(i & 0x3fff) << 12) + mem->sys_start;
    mem->card_start &= 0x3ffffff;
    return 0;
} /* i365_get_mem_map */

/*====================================================================*/
  
static int i365_set_mem_map(u_short sock, struct pcmcia_mem_map *mem)
{
    u_short base, i;
    u_char map;
    
    map = mem->map;
    if ((map > 4) || (mem->card_start > 0x3ffffff) ||
	(mem->sys_start > 0xffffff) || (mem->sys_stop > 0xffffff) ||
	(mem->sys_start > mem->sys_stop) || (mem->extra_ws > 3))
	return -EINVAL;
    /* Turn off the window before changing anything */
    if (i365_get(sock, I365_ADDRWIN) & I365_ENA_MEM(map))
	i365_bclr(sock, I365_ADDRWIN, I365_ENA_MEM(map));
    base = I365_MEM(map);
    
    i = mem->sys_start >> 12;
    if (mem->flags & MAP_16BIT) i |= I365_MEM_16BIT;
    if (mem->flags & MAP_0WS) i |= I365_MEM_0WS;
    i365_set_pair(sock, base+I365_W_START, i);
    
    i = mem->sys_stop >> 12;
    if (mem->extra_ws & 1) i |= I365_MEM_WS0;
    if (mem->extra_ws & 2) i |= I365_MEM_WS1;
    i365_set_pair(sock, base+I365_W_STOP, i);
    
    i = ((mem->card_start - mem->sys_start) >> 12) & 0x3fff;
    if (mem->flags & MAP_WRPROT) i |= I365_MEM_WRPROT;
    if (mem->flags & MAP_ATTRIB) i |= I365_MEM_REG;
    i365_set_pair(sock, base+I365_W_OFF, i);
    
    /* Turn on the window if necessary */
    if (mem->flags & MAP_ACTIVE)
	i365_bset(sock, I365_ADDRWIN, I365_ENA_MEM(map));
    return 0;
} /* i365_set_mem_map */

/*====================================================================*/

typedef int (*subfn_t)(u_short, void *);

static subfn_t ioctl_table[] = {
    (subfn_t)&i365_get_status,
    (subfn_t)&i365_get_flags,
    (subfn_t)&i365_set_flags,
    (subfn_t)&i365_csc_on,
    (subfn_t)&i365_csc_off,
    (subfn_t)&i365_csc_poll,
    (subfn_t)&i365_get_irq,
    (subfn_t)&i365_set_irq,
    (subfn_t)&i365_get_io_map,
    (subfn_t)&i365_set_io_map,
    (subfn_t)&i365_get_mem_map,
    (subfn_t)&i365_set_mem_map,
    (subfn_t)&i365_reset,
};

#define NFUNC 13

static int i365_ioctl(struct pcmcia_ctrl_t *ctrl, u_int cmd, void *arg)
{
    int err, sock = ctrl->sock;

    cmd = cmd & 0xff;
    if (cmd < NFUNC)
	err = ioctl_table[cmd](sock, arg);
    else
	err = -EINVAL;
    
    if (pc_debug) 
	printk("i365_ioctl(%d, %d, 0x%p) = %d\n",
	       sock, cmd, arg, err);
    
    return err;
} /* i365_ioctl */

/*====================================================================*/

#ifdef MODULE
char kernel_version[] = UTS_RELEASE;

int init_module(void)
{
    return i365_init();
}

void cleanup_module(void)
{
    i365_finish();
}
#endif /* MODULE */
