/*======================================================================

    A driver for PCMCIA serial devices

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/tty.h>
#include <linux/serial.h>
#include <asm/io.h>
#include <asm/system.h>

#ifdef MODULE
#include <linux/module.h>
#include "../../tools/version.h"
#endif

#include "cs_types.h"
#include "cs.h"
#include "cistpl.h"
#include "ds.h"

#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
static char *version =
    "serial_cs.c v$Revision: 1.8 $: $Date: 1994/06/22 07:08:07 $ (David Hinds)\n";
#endif

/*====================================================================*/

static void serial_config(u_long arg);
static void serial_release(u_long arg);
static int serial_event(event_t event, int priority,
			event_callback_args_t *args);

static dev_info_t dev_info = "serial_cs";

static dev_link_t *serial_attach(void);
static void serial_detach(dev_link_t *);

static dev_link_t default_link = {
    "\0",
    0,
    { 0, 0, 0, 0, &serial_config },
    { 0, 0, 0, 0, &serial_release },
    NULL,
    { 0, 0, 8, 0, 0, 0, 0, 3 },
    { 0, IRQ_TYPE_EXCLUSIVE, 0, IRQ_INFO2_VALID|IRQ_LEVEL_ID, 0x86bc },
    {
	0, CONF_ENABLE_IRQ, 50, 0, 0, INT_MEMORY_AND_IO,
	0, 0, 0, 0, 0, PRESENT_OPTION
    },
    NULL,
    NULL
};

static dev_link_t *dev_list = NULL;

/*======================================================================

    If we're linked into the kernel, this gets called from init.c.
    
======================================================================*/

#ifndef MODULE
u_long serial_init(u_long kmem_start, u_long kmem_end)
{
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk(version);
#endif
    register_pcmcia_driver(&dev_info,
			   (void *(*)(void))&serial_attach,
			   (void (*)(void *))&serial_detach);
    return kmem_start;
}
#endif

/*======================================================================

    serial_attach() creates an "instance" of the driver, allocating
    local data structures for one device.  The device is registered
    with Card Services.

======================================================================*/

static dev_link_t *serial_attach(void)
{
    client_reg_t client_reg;
    dev_link_t *link;
    int ret;
    
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("serial_attach()\n");
#endif

#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif

    /* Create new serial device */
    link = kmalloc(sizeof(struct dev_link_t), GFP_KERNEL);
    *link = default_link;
    link->config.data = (u_long)link;
    link->release.data = (u_long)link;
    link->priv = kmalloc(sizeof(int), GFP_KERNEL);
    *(int *)link->priv = -1;
    
    /* Register with Card Services */
    link->next = dev_list;
    dev_list = link;
    client_reg.dev_info = &dev_info;
    client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
    client_reg.EventMask = EVENT_CARD_DETECT;
    client_reg.event_handler = &serial_event;
    client_reg.Version = 0x0210;
    client_reg.Socket = 0;
    ret = CardServices(RegisterClient, &link->handle, &client_reg);
    if (ret != CS_SUCCESS) {
	printk("serial_cs: RegisterClient failed, ret = %d\n", ret);
	serial_detach(link);
	return NULL;
    }
    
    return link;
} /* serial_attach */

/*======================================================================

    This deletes a driver "instance".  The device is de-registered
    with Card Services.  If it has been released, all local data
    structures are freed.  Otherwise, the structures will be freed
    when the device is released.

======================================================================*/

static void serial_detach(dev_link_t *link)
{
    dev_link_t **linkp;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("serial_detach(0x%p)\n", link);
#endif
    
    /* Locate device structure */
    for (linkp = &dev_list; *linkp; linkp = &(*linkp)->next)
	if (*linkp == link) break;
    if (*linkp == NULL)
	return;

    if (link->state & DEV_CONFIG) {
	printk("serial_cs: detach postponed, '%s' still locked\n",
	       link->dev_name);
	link->state |= DEV_STALE_LINK;
	return;
    }
    
    if (link->handle)
	CardServices(DeregisterClient, link->handle);
    
    /* Unlink device structure, free bits */
    *linkp = link->next;
    kfree_s(link->priv, sizeof(int));
    kfree_s(link, sizeof(struct dev_link_t));
    
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
    
} /* serial_detach */

/*======================================================================

    serial_config() is scheduled to run after a CARD_INSERTION event
    is received, to configure the PCMCIA socket, and to make the
    serial device available to the system.

======================================================================*/

void serial_config(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;
    client_handle_t handle;
    tuple_t tuple;
    u_char tuple_data[32];
    cisparse_t parse;
    struct serial_struct serial;
    int i, line;

    sti();
    handle = link->handle;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("serial_config(0x%p)\n", link);
#endif
    
    do {
	tuple.DesiredTuple = CISTPL_CONFIG;
	i = CardServices(GetFirstTuple, handle, &tuple);
	if (i != CS_SUCCESS) break;
	tuple.TupleOffset = 0;
	tuple.TupleDataMax = 32;
	tuple.TupleData = tuple_data;
	i = CardServices(GetTupleData, handle, &tuple);
	if (i != CS_SUCCESS) break;
	i = CardServices(ParseTuple, handle, &tuple, &parse);
	if (i != CS_SUCCESS) break;
	link->conf.ConfigBase = (caddr_t)parse.config.base;
	/* We assume the actual config index is irrelevant */
	link->conf.ConfigIndex = parse.config.last_idx;
    } while (0);
    
    if (i != 0) {
	printk("serial_cs: unable to parse CIS\n");
	link->state &= ~DEV_CONFIG_PENDING;
	return;
    }
    
    /* Configure card */
    link->state |= DEV_CONFIG;

#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif

    do {
	
	i = CardServices(RequestIO, link->handle, &link->io);
	if (i != CS_SUCCESS) {
	    printk("serial_cs: RequestIO failed, ret = %d\n", i);
	    break;
	}
	i = CardServices(RequestIRQ, link->handle, &link->irq);
	if (i != CS_SUCCESS) {
	    printk("serial_cs: RequestIRQ failed, ret = %d\n", i);
	    break;
	}
	i = CardServices(RequestConfiguration, link->handle, &link->conf);
	if (i != CS_SUCCESS) {
	    printk("serial_cs: RequestConfiguration failed, ret = %d\n", i);
	    break;
	}
	serial.port = link->io.BasePort1;
	serial.irq = link->irq.AssignedIRQ;
	line = register_serial(&serial);
	if (line < 0) {
	    printk("serial_cs: register_serial() failed\n");
	    i = -1;
	    break;
	}
	*(int *)link->priv = line;
	sprintf(link->dev_name, "ttyS%d", line);
	    
    } while (0);

    link->state &= ~DEV_CONFIG_PENDING;
    if (i != 0)
	serial_release((u_long)link);

} /* serial_config */

/*======================================================================

    After a card is removed, serial_release() will unregister the net
    device, and release the PCMCIA configuration.  If the device is
    still open, this will be postponed until it is closed.
    
======================================================================*/

void serial_release(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;
    sti();
    
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("serial_release(0x%p)\n", link);
#endif

    if (*(int *)link->priv >= 0) {
	link->dev_name[0] = '\0';
	unregister_serial(*(int *)link->priv);
    }
    
    CardServices(ReleaseConfiguration, link->handle);
    CardServices(ReleaseIO, link->handle, &link->io);
    CardServices(ReleaseIRQ, link->handle, &link->irq);
    
    link->state &= ~DEV_CONFIG;
    if (link->state & DEV_STALE_LINK)
	serial_detach(link);
    
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
} /* serial_release */

/*======================================================================

    The card status event handler.  Mostly, this schedules other
    stuff to run after an event is received.  A CARD_REMOVAL event
    also sets some flags to discourage the serial drivers from
    talking to the ports.
    
======================================================================*/

int serial_event(event_t event, int priority,
		event_callback_args_t *args)
{
    dev_link_t *link;

#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk("serial_event()\n");
#endif
    
    /* Locate device structure */
    for (link = dev_list; link; link = link->next)
	if (link->handle == args->client_handle) break;
    if (link == NULL) {
	printk("serial_cs: bad event handle 0x%p!\n", args->client_handle);
	return 0;
    }
    
    switch (event) {
#ifdef PCMCIA_DEBUG
    case CS_EVENT_REGISTRATION_COMPLETE:
	if (pc_debug)
	    printk("serial_cs: registration complete\n");
	break;
#endif
    case CS_EVENT_CARD_REMOVAL:
	link->state &= ~DEV_PRESENT;
	if (link->state & DEV_CONFIG) {
	    link->release.expires = 5;
	    add_timer(&link->release);
	}
	break;
    case CS_EVENT_CARD_INSERTION:
	link->state |= DEV_PRESENT | DEV_CONFIG_PENDING;
	link->config.expires = 5;
	add_timer(&link->config);
	break;
    case CS_EVENT_RESET_REQUEST:
	return 1;
	break;
    }
    return 0;
} /* serial_event */

/*====================================================================*/

#ifdef MODULE
char kernel_version[] = UTS_RELEASE;

int init_module(void)
{
#ifdef PCMCIA_DEBUG
    if (pc_debug)
	printk(version);
#endif
    register_pcmcia_driver(&dev_info, &serial_attach, &serial_detach);
    return 0;
}

void cleanup_module(void)
{
    printk("serial_cs: unloading\n");
}
#endif /* MODULE */
