/*======================================================================

    Device driver for Databook TCIC/2 PCMCIA controller

    Written by David Hinds, dhinds@allegro.stanford.edu
    
======================================================================*/

#include <sys/types.h>
#include <linux/fcntl.h>
#include <linux/string.h>

#include <asm/io.h>
#include <asm/bitops.h>
#include <asm/segment.h>
#include <asm/system.h>

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/ioport.h>
#include <linux/malloc.h>

#ifdef MODULE
#include <linux/module.h>
#include "../../tools/version.h"
#endif

#ifdef PCMCIA_DEBUG
int pc_debug = PCMCIA_DEBUG;
#else
int pc_debug = 0;
#endif

#include <linux/pcmcia.h>
#include "tcic.h"

static const char *version =
    "tcic.c v$Revision: 1.2 $: $Date: 1994/06/24 16:23:49 $ (David Hinds)\n";

/*====================================================================*/

static int tcic_open(struct pcmcia_ctrl_t *ctrl);
static void tcic_close(struct pcmcia_ctrl_t *ctrl);
static int tcic_ioctl(struct pcmcia_ctrl_t *ctrl, u_int cmd, void *arg);

static pcmcia_ctrl_t tcic_ctrl =
    { "TCIC/2", 0, 0, 0, &tcic_open, &tcic_close, &tcic_ioctl, NULL };

static pcmcia_ctrl_t *tcic_list = NULL;

static resource_entry_t tcic_ports = { TCIC_BASE, 16, NULL };

static u_char last_sstat[2];

/*====================================================================*/

#ifdef PCMCIA_DEBUG_X
static u_char tcic_getb(u_char reg)
{
    u_char val = inb(TCIC_BASE+reg);
    printk("tcic_getb(%#x) = %#x\n", TCIC_BASE+reg, val);
    return val;
}

static u_short tcic_getw(u_char reg)
{
    u_short val = inw(TCIC_BASE+reg);
    printk("tcic_getw(%#x) = %#x\n", TCIC_BASE+reg, val);
    return val;
}

static void tcic_setb(u_char reg, u_char data)
{
    printk("tcic_setb(%#x, %#x)\n", TCIC_BASE+reg, data);
    outb(data, TCIC_BASE+reg);
}

static void tcic_setw(u_char reg, u_short data)
{
    printk("tcic_setw(%#x, %#x)\n", TCIC_BASE+reg, data);
    outw(data, TCIC_BASE+reg);
}
#else
#define tcic_getb(reg) inb(TCIC_BASE+reg)
#define tcic_getw(reg) inw(TCIC_BASE+reg)
#define tcic_setb(reg, data) outb(data, TCIC_BASE+reg)
#define tcic_setw(reg, data) outw(data, TCIC_BASE+reg)
#endif

static void tcic_setl(u_char reg, u_long data)
{
#ifdef PCMCIA_DEBUG_X
    printk("tcic_setl(%#x, %#lx)\n", TCIC_BASE+reg, data);
#endif
    outw(data & 0xffff, TCIC_BASE+reg);
    outw(data >> 16, TCIC_BASE+reg+2);
}

static u_char tcic_aux_getb(u_short reg)
{
    u_char mode = (tcic_getb(TCIC_MODE) & TCIC_MODE_PGMMASK) | reg;
    tcic_setb(TCIC_MODE, mode);
    return tcic_getb(TCIC_AUX);
}

static void tcic_aux_setb(u_short reg, u_char data)
{
    u_char mode = (tcic_getb(TCIC_MODE) & TCIC_MODE_PGMMASK) | reg;
    tcic_setb(TCIC_MODE, mode);
    tcic_setb(TCIC_AUX, data);
}

static u_short tcic_aux_getw(u_short reg)
{
    u_char mode = (tcic_getb(TCIC_MODE) & TCIC_MODE_PGMMASK) | reg;
    tcic_setb(TCIC_MODE, mode);
    return tcic_getw(TCIC_AUX);
}

static void tcic_aux_setw(u_short reg, u_short data)
{
    u_char mode = (tcic_getb(TCIC_MODE) & TCIC_MODE_PGMMASK) | reg;
    tcic_setb(TCIC_MODE, mode);
    tcic_setw(TCIC_AUX, data);
}

/*====================================================================*/

int tcic_init(void)
{
    int irqlist[] = { 14, 10, 7, 6, 5, 4, 3, 0 };
    int i, sock, *irqp;
    pcmcia_ctrl_t *ctrl;

    if (pc_debug)
	printk(version);
    
    printk("Databook TCIC/2 PCMCIA probe: ");
    sock = 0;
    if (check_io_region(&tcic_ports) == 0) {
	tcic_setw(TCIC_ADDR, 0);
	if (tcic_getw(TCIC_ADDR) == 0) {
	    tcic_setw(TCIC_ADDR, 0xc3a5);
	    if (tcic_getw(TCIC_ADDR) == 0xc3a5) sock = 2;
	}
    }
    else
	printk("could not allocate ports, ");
    
    if (sock == 0) {
	printk("not found.\n");
	return -ENODEV;
    }
    
    alloc_io_region(&tcic_ports);

    for (irqp = irqlist; *irqp != 0; irqp++) {
	if (request_irq(*irqp, NULL) != -EBUSY) break;
    }
    
    printk("%d sockets, ", sock);
    if (*irqp != 0)
	printk("status change on IRQ %d\n", *irqp);
    else
	printk("no status change irq\n");

    tcic_aux_setb(TCIC_AUX_WCTL, TCIC_WAIT_SENSE|TCIC_WAIT_ASYNC|7);
    tcic_aux_setw(TCIC_AUX_SYSCFG, TCIC_SYSCFG_AUTOBUSY|0x0a00);
    for (i = 0; i < sock; i++) {
	ctrl = kmalloc(sizeof(pcmcia_ctrl_t), GFP_KERNEL);
	*ctrl = tcic_ctrl;
	ctrl->sock = i;
	ctrl->csc_irq = *irqp;
	if (register_pcmcia(ctrl) != 0) {
	    printk("tcic: register_pcmcia() failed, sock %d\n", i);
	    kfree_s(ctrl, sizeof(pcmcia_ctrl_t));
	    break;
	}
	tcic_aux_setb(TCIC_AUX_ILOCK, TCIC_ILOCK_HOLD_CCLK |
		      TCIC_ILOCK_CWAIT | TCIC_ILOCK_CRESENA);
	last_sstat[i] = tcic_getb(TCIC_SSTAT);
	ctrl->next = tcic_list;
	tcic_list = ctrl;
    }

    if (tcic_list == NULL) {
	free_io_region(&tcic_ports);
	return -ENODEV;
    }
    else
	return 0;
    
} /* tcic_init */

/*====================================================================*/

static void tcic_finish(void)
{
    pcmcia_ctrl_t *c, *d;

    for (c = tcic_list; c; c = d) {
	unregister_pcmcia(c);
	d = c->next;
	kfree_s(c, sizeof(pcmcia_ctrl_t));
    }
    free_io_region(&tcic_ports);
} /* tcic_finish */

/*====================================================================*/

static int tcic_open(struct pcmcia_ctrl_t *ctrl)
{
#ifdef MODULE
    MOD_INC_USE_COUNT;
#endif
    return 0;
}

static void tcic_close(struct pcmcia_ctrl_t *ctrl)
{
#ifdef MODULE
    MOD_DEC_USE_COUNT;
#endif
}

/*====================================================================*/

static void busy_loop(u_long len)
{
    u_long timeout = jiffies + len;
    u_long flags;
    save_flags(flags);
    sti();
    while (timeout >= jiffies)
	;
    restore_flags(flags);
} /* busy_loop */

/*====================================================================*/

static int tcic_get_status(u_short sock, u_int *value)
{
    u_char reg;

    tcic_setl(TCIC_ADDR, (sock << TCIC_ADDR_SS_SHFT)
	      | TCIC_ADDR_INDREG | TCIC_SCF1(sock));
    reg = tcic_getb(TCIC_SSTAT);
    *value  = (reg & TCIC_SSTAT_CD) ? PC_DETECT : 0;
    *value |= (reg & TCIC_SSTAT_WP) ? PC_WRPROT : 0;
    if (tcic_getw(TCIC_DATA) & TCIC_SCF1_IOSTS) {
	*value |= (reg & TCIC_SSTAT_LBAT1) ? PC_STSCHG : 0;
    }
    else {
	*value |= (reg & TCIC_SSTAT_RDY) ? PC_READY : 0;
	*value |= (reg & TCIC_SSTAT_LBAT1) ? PC_BATDEAD : 0;
	*value |= (reg & TCIC_SSTAT_LBAT2) ? PC_BATWARN : 0;
    }
    reg = tcic_getb(TCIC_PWR);
    if (reg & (TCIC_PWR_VCC(sock)|TCIC_PWR_VPP(sock)))
	*value |= PC_POWERON;
    return 0;
} /* tcic_get_status */
  
/*====================================================================*/

static int tcic_get_flags(u_short sock, u_int *value)
{
    u_char reg;
    
    tcic_setl(TCIC_ADDR, (sock << TCIC_ADDR_SS_SHFT)
	      | TCIC_ADDR_INDREG | TCIC_SCF1(sock));
    *value = (tcic_getw(TCIC_DATA) & TCIC_SCF1_IOSTS) ? PC_IOCARD : 0;

    reg = tcic_getb(TCIC_PWR);
    if (reg & TCIC_PWR_VCC(sock))
	if (reg & TCIC_PWR_VPP(sock))
	    *value |= PC_VCC_5V;
	else
	    *value |= PC_VCC_5V | PC_VPP_5V;
    else
	if (reg & TCIC_PWR_VPP(sock))
	    *value |= PC_VCC_5V | PC_VPP_12V;
    reg = tcic_aux_getb(TCIC_AUX_ILOCK);
    *value |= (reg & TCIC_ILOCK_CRESET) ? PC_RESET : 0;
    
    return 0;
} /* tcic_get_flags */

/*====================================================================*/

static int tcic_set_flags(u_short sock, u_int *value)
{
    u_char reg;
    u_short scf1;

    tcic_setl(TCIC_ADDR, (sock << TCIC_ADDR_SS_SHFT)
	      | TCIC_ADDR_INDREG | TCIC_SCF1(sock));

    reg = tcic_getb(TCIC_PWR);
    reg &= ~(TCIC_PWR_VCC(sock) | TCIC_PWR_VPP(sock));
    if (*value & PC_VCC_5V) {
	if (*value & PC_VPP_12V)
	    reg |= TCIC_PWR_VPP(sock);
	else if (*value & PC_VPP_5V)
	    reg |= TCIC_PWR_VCC(sock);
	else
	    reg |= TCIC_PWR_VCC(sock) | TCIC_PWR_VPP(sock);
    }
    tcic_setb(TCIC_PWR, reg);
    busy_loop(2);

    scf1 = tcic_getw(TCIC_DATA);
    if (*value & PC_IOCARD)
	scf1 |= TCIC_SCF1_IOSTS | TCIC_SCF1_SPKR;
    else
	scf1 &= ~(TCIC_SCF1_IOSTS | TCIC_SCF1_SPKR);
    tcic_setw(TCIC_DATA, scf1);

    if (*value & PC_VCC_MASK)
	tcic_setb(TCIC_SCTRL, TCIC_SCTRL_ENA);
    reg = tcic_aux_getb(TCIC_AUX_ILOCK) & ~TCIC_ILOCK_CRESET;
    if (*value & PC_RESET)
	reg |= TCIC_ILOCK_CRESET;
    tcic_aux_setb(TCIC_AUX_ILOCK, reg);
    
    return 0;
} /* tcic_set_flags */
  
/*====================================================================*/

static int tcic_csc_on(u_short sock, int *irq)
{
    u_char reg;
    u_short scf2;

    /* Turn on all card status change interrupts */
    tcic_setl(TCIC_ADDR, (sock << TCIC_ADDR_SS_SHFT)
	      | TCIC_ADDR_INDREG | TCIC_SCF2(sock));
    reg = tcic_aux_getb(TCIC_AUX_SYSCFG);
    tcic_aux_setb(TCIC_AUX_SYSCFG, reg | *irq);
    /* Donald's code sets the status irq to open drain ??? */
    tcic_setb(TCIC_IENA, TCIC_IENA_CDCHG | TCIC_IENA_CFG_HIGH);
    scf2 = tcic_getw(TCIC_DATA);
    scf2 &= TCIC_SCF2_MDBR | TCIC_SCF2_IDBR | TCIC_SCF2_RI;
    tcic_setw(TCIC_DATA, scf2);
    /* flush current card status changes */
    tcic_setb(TCIC_ICSR, TCIC_ICSR_CLEAR);
    return 0;
} /* tcic_csc_on */

/*====================================================================*/

static int tcic_csc_off(u_short sock, void *arg)
{
    /* Turn off all card status change interrupts */
    tcic_setw(TCIC_ADDR+2, (sock << TCIC_SS_SHFT));
    tcic_setb(TCIC_IENA, TCIC_IENA_CFG_OFF);
    return 0;
} /* tcic_csc_off */

/*====================================================================*/

static int tcic_csc_poll(u_short sock, int *ret)
{
    u_char latch, sstat;
    
    tcic_setl(TCIC_ADDR, (sock << TCIC_ADDR_SS_SHFT)
	      | TCIC_ADDR_INDREG | TCIC_SCF1(sock));
    /* Anything to report? */
/*    if (!(tcic_getb(TCIC_ICSR) & TCIC_ICSR_CDCHG)) {
	*ret = 0;
	return 0;
    }*/
/*    latch = tcic_aux_getw(TCIC_AUX_WCTL) >> 8;*/
    sstat = tcic_getb(TCIC_SSTAT);
/*    latch = latch ^ sstat;*/
    latch = sstat ^ last_sstat[sock];
    last_sstat[sock] = sstat;
    tcic_setb(TCIC_ICSR, TCIC_ICSR_CLEAR);
    
    *ret  = (latch & TCIC_SSTAT_CD) ? PC_DETECT : 0;
    *ret |= (latch & TCIC_SSTAT_WP) ? PC_WRPROT : 0;
    if (tcic_getw(TCIC_DATA) & TCIC_SCF1_IOSTS) {
	*ret |= (latch & TCIC_SSTAT_LBAT1) ? PC_STSCHG : 0;
    }
    else {
	*ret |= (latch & TCIC_SSTAT_RDY) ? PC_READY : 0;
	*ret |= (latch & TCIC_SSTAT_LBAT1) ? PC_BATDEAD : 0;
	*ret |= (latch & TCIC_SSTAT_LBAT2) ? PC_BATWARN : 0;
    }
    return 0;
} /* tcic_csc_poll */

/*====================================================================*/

static int tcic_get_irq(u_short sock, u_int *value)
{
    tcic_setl(TCIC_ADDR, (sock << TCIC_ADDR_SS_SHFT)
	      | TCIC_ADDR_INDREG | TCIC_SCF1(sock));
    *value = tcic_getw(TCIC_DATA) & 0x0f;
    return 0;
} /* tcic_get_irq */
  
/*====================================================================*/

static int tcic_set_irq(u_short sock, u_int *value)
{
    u_short scf1;
    
    if (*value > 15) return -EINVAL;
    tcic_setl(TCIC_ADDR, (sock << TCIC_ADDR_SS_SHFT)
	      | TCIC_ADDR_INDREG | TCIC_SCF1(sock));
    
    scf1 = tcic_getw(TCIC_DATA);
    /* Donald's code sets the irq as active high ??? */
    scf1 &= ~(TCIC_SCF1_IRQOC | TCIC_SCF1_IRQ_MASK);
    scf1 |= *value;
    tcic_setw(TCIC_DATA, scf1);
    
    return 0;
} /* tcic_set_irq */

/*====================================================================*/

static int tcic_reset(u_short sock, void *arg)
{
    int i;
    u_short reg;
    
    /* Turn off interrupts, select memory-only interface */
    tcic_setl(TCIC_ADDR, (sock << TCIC_ADDR_SS_SHFT)
	      | TCIC_ADDR_INDREG | TCIC_SCF1(sock));
    reg = tcic_getw(TCIC_DATA);
    reg &= ~(TCIC_SCF1_IRQ_MASK | TCIC_SCF1_IOSTS);
    tcic_setw(TCIC_DATA, reg);
    
    /* Turn off I/O windows */
    for (i = 0; i < 2; i++) {
	tcic_setw(TCIC_ADDR, TCIC_IWIN(sock, i) + TCIC_ICTL_X);
	tcic_setw(TCIC_DATA, 0);
    }
    
    /* Turn off memory windows */
    for (i = 0; i < 4; i++) {
	tcic_setw(TCIC_ADDR, TCIC_MWIN(sock, i) + TCIC_MCTL_X);
	tcic_setw(TCIC_DATA, 0);
    }
    
    return 0;
} /* tcic_reset */

/*====================================================================*/

static int tcic_get_io_map(u_short sock, struct pcmcia_io_map *io)
{
    u_short base, ioctl;
    u_long addr;
    
    if (io->map > 1) return -EINVAL;
    tcic_setw(TCIC_ADDR+2, TCIC_ADR2_INDREG | (sock << TCIC_SS_SHFT));
    addr = TCIC_IWIN(sock, io->map);
    tcic_setw(TCIC_ADDR, addr + TCIC_IBASE_X);
    base = tcic_getw(TCIC_DATA);
    tcic_setl(TCIC_ADDR, addr + TCIC_ICTL_X);
    ioctl = tcic_getw(TCIC_DATA);

    if (ioctl & TCIC_ICTL_TINY)
	io->start = io->stop = base;
    else {
	io->start = base & (base-1);
	io->stop = io->start + (base ^ (base-1));
    }
    io->extra_ws = ioctl & TCIC_ICTL_WSCNT_MASK;
    io->flags  = (ioctl & TCIC_ICTL_ENA) ? MAP_ACTIVE : 0;
    if (ioctl & TCIC_ICTL_BW_MASK == TCIC_ICTL_B16)
	io->flags |= MAP_16BIT;
    return 0;
} /* tcic_get_io_map */

/*====================================================================*/

static int tcic_set_io_map(u_short sock, struct pcmcia_io_map *io)
{
    u_long addr;
    u_short base, len, ioctl;
    
    if ((io->map > 1) || (io->start > 0xffff) || (io->stop > 0xffff) ||
	(io->stop < io->start)) return -EINVAL;
    tcic_setw(TCIC_ADDR+2, TCIC_ADR2_INDREG | (sock << TCIC_SS_SHFT));
    addr = TCIC_IWIN(sock, io->map);

    base = io->start; len = io->stop - io->start;
    /* Check to see that len+1 is power of two, etc */
    if ((len & (len+1)) || (base & len)) return -EINVAL;
    base |= (len+1)>>1;
    tcic_setw(TCIC_ADDR, addr + TCIC_IBASE_X);
    tcic_setw(TCIC_DATA, base);
    
    ioctl  = TCIC_ICTL_QUIET | (sock << TCIC_ICTL_SS_SHFT);
    ioctl |= (len == 0) ? TCIC_ICTL_TINY : 0;
    ioctl |= (io->flags & MAP_ACTIVE) ? TCIC_ICTL_ENA : 0;
    ioctl |= io->extra_ws & TCIC_ICTL_WSCNT_MASK;
    tcic_setw(TCIC_ADDR, addr + TCIC_ICTL_X);
    tcic_setw(TCIC_DATA, ioctl);
    
    return 0;
} /* tcic_set_io_map */

/*====================================================================*/

static int tcic_get_mem_map(u_short sock, struct pcmcia_mem_map *mem)
{
    u_short addr, ctl;
    u_long base, mmap;
    
    if (mem->map > 3) return -EINVAL;
    tcic_setw(TCIC_ADDR+2, TCIC_ADR2_INDREG | (sock << TCIC_SS_SHFT));
    addr = TCIC_MWIN(sock, mem->map);
    
    tcic_setw(TCIC_ADDR, addr + TCIC_MBASE_X);
    base = tcic_getw(TCIC_DATA);
    if (base & TCIC_MBASE_4K_BIT) {
	mem->sys_start = base & TCIC_MBASE_HA_MASK;
	mem->sys_stop = mem->sys_start;
    }
    else {
	base &= TCIC_MBASE_HA_MASK;
	mem->sys_start = (base & (base-1));
	mem->sys_stop = mem->sys_start + (base ^ (base-1));
    }
    mem->sys_start = mem->sys_start << TCIC_MBASE_HA_SHFT;
    mem->sys_stop = (mem->sys_stop << TCIC_MBASE_HA_SHFT) + 0x0fff;
    
    tcic_setw(TCIC_ADDR, addr + TCIC_MMAP_X);
    mmap = tcic_getw(TCIC_DATA);
    mem->flags = (mmap & TCIC_MMAP_REG) ? MAP_ATTRIB : 0;
    mmap &= TCIC_MMAP_CA_MASK;
    mem->card_start = mem->sys_start + (mmap << TCIC_MMAP_CA_SHFT);
    
    tcic_setw(TCIC_ADDR, addr + TCIC_MCTL_X);
    ctl = tcic_getw(TCIC_DATA);
    mem->flags |= (ctl & TCIC_MCTL_ENA) ? MAP_ACTIVE : 0;
    mem->flags |= (ctl & TCIC_MCTL_B8) ? 0 : MAP_16BIT;
    mem->flags |= (ctl & TCIC_MCTL_WP) ? MAP_WRPROT : 0;
    mem->extra_ws = (ctl & TCIC_MCTL_WSCNT_MASK);
    
    return 0;
} /* tcic_get_mem_map */

/*====================================================================*/
  
static int tcic_set_mem_map(u_short sock, struct pcmcia_mem_map *mem)
{
    u_short addr, ctl;
    u_long base, len, mmap;
    
    if ((mem->map > 3) || (mem->card_start > 0xffffff) ||
	(mem->sys_start > 0xffffff) || (mem->sys_stop > 0xffffff) ||
	(mem->sys_start > mem->sys_stop) || (mem->extra_ws > 7))
	return -EINVAL;
    tcic_setw(TCIC_ADDR+2, TCIC_ADR2_INDREG | (sock << TCIC_SS_SHFT));
    addr = TCIC_MWIN(sock, mem->map);

    base = mem->sys_start; len = mem->sys_stop - mem->sys_start;
    if ((len & (len+1)) || (base & len)) return -EINVAL;
    if (len == 0x0fff)
	base = (base >> TCIC_MBASE_HA_SHFT) | TCIC_MBASE_4K_BIT;
    else
	base = (base | (len+1)>>1) >> TCIC_MBASE_HA_SHFT;
    tcic_setw(TCIC_ADDR, addr + TCIC_MBASE_X);
    tcic_setw(TCIC_DATA, base);
    
    mmap = mem->card_start - mem->sys_start;
    mmap = (mmap >> TCIC_MMAP_CA_SHFT) & TCIC_MMAP_CA_MASK;
    if (mem->flags & MAP_ATTRIB) mmap |= TCIC_MMAP_REG;
    tcic_setw(TCIC_ADDR, addr + TCIC_MMAP_X);
    tcic_setw(TCIC_DATA, mmap);

    ctl  = TCIC_MCTL_QUIET | (sock << TCIC_MCTL_SS_SHFT);
    ctl |= mem->extra_ws & TCIC_MCTL_WSCNT_MASK;
    ctl |= (mem->flags & MAP_16BIT) ? 0 : TCIC_MCTL_B8;
    ctl |= (mem->flags & MAP_WRPROT) ? TCIC_MCTL_WP : 0;
    ctl |= (mem->flags & MAP_ACTIVE) ? TCIC_MCTL_ENA : 0;
    tcic_setw(TCIC_ADDR, addr + TCIC_MCTL_X);
    tcic_setw(TCIC_DATA, ctl);
    
    return 0;
} /* tcic_set_mem_map */

/*====================================================================*/

typedef int (*subfn_t)(u_short, void *);
    
static subfn_t ioctl_table[] = {
    (subfn_t)&tcic_get_status,
    (subfn_t)&tcic_get_flags,
    (subfn_t)&tcic_set_flags,
    (subfn_t)&tcic_csc_on,
    (subfn_t)&tcic_csc_off,
    (subfn_t)&tcic_csc_poll,
    (subfn_t)&tcic_get_irq,
    (subfn_t)&tcic_set_irq,
    (subfn_t)&tcic_get_io_map,
    (subfn_t)&tcic_set_io_map,
    (subfn_t)&tcic_get_mem_map,
    (subfn_t)&tcic_set_mem_map,
    (subfn_t)&tcic_reset,
};
#define NFUNC 13

static int tcic_ioctl(struct pcmcia_ctrl_t *ctrl, u_int cmd, void *arg)
{
    int err, sock = ctrl->sock;

    cmd = cmd & 0xff;
    if (cmd < NFUNC)
	err = ioctl_table[cmd](sock, arg);
    else
	err = -EINVAL;
    
    if (pc_debug) 
	printk("tcic_ioctl(%d, %x, 0x%p) = %d\n", sock, cmd, arg, err);
    return err;
} /* tcic_ioctl */

/*====================================================================*/

#ifdef MODULE
char kernel_version[] = UTS_RELEASE;

int init_module(void)
{
    return tcic_init();
}

void cleanup_module(void)
{
    tcic_finish();
}
#endif /* MODULE */
